#!/usr/bin/env python
# coding: utf-8

import pytest
import capytaine as cpt

try:
    import vtk
except ImportError:
    vtk = None

try:
    import IPython
except ImportError:
    IPython = None

@pytest.mark.skipif(vtk is None or IPython is None,
                    reason='vtk and/or Ipython are not installed')
def test_animation_of_dofs():
    body = cpt.Sphere()
    body.add_translation_dof(name="Heave")
    animation = body.animate({"Heave": 0.2}, loop_duration=1.0)
    animation.embed_in_notebook()

@pytest.mark.skipif(vtk is None, reason='vtk is not installed')
def test_animate_missing_dof():
    body = cpt.Sphere()
    with pytest.raises(ValueError, match=".*no dof of this name.*"):
        body.animate({"Heave": 0.1}, loop_duration=1.0)



"""This module implements block matrices to be used in Hierarchical Toeplitz matrices.

It takes inspiration from the following works:

* `openHmx module from Gypsilab by Matthieu Aussal (GPL licensed) <https://github.com/matthieuaussal/gypsilab>`_
* `HierarchicalMatrices by Markus Neumann (GPL licensed) <https://github.com/maekke97/HierarchicalMatrices>`_
"""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging

from numbers import Number
from typing import Tuple, List, Callable, Union, Iterable
from itertools import cycle, accumulate, chain, product
from collections.abc import Iterator

import numpy as np

from capytaine.matrices.low_rank import LowRankMatrix
from capytaine.tools.optional_imports import import_optional_dependency

LOG = logging.getLogger(__name__)


class BlockMatrix:
    """A (2D) matrix, stored as a set of submatrices (or blocks).

    Parameters
    ----------
    blocks: list of list of matrices
        The blocks of the block matrix.
    check: bool, optional
        Should the code perform sanity checks on the inputs? (default: True)

    Attributes
    ----------
    shape: pair of ints
        shape of the full matrix
    nb_blocks: pair of ints
        number of blocks in each directions.
        Example::

            AAAABB
            AAAABB  ->  nb_blocks = (1, 2)
            AAAABB
    """

    ndim = 2  # Other dimensions have not been implemented.

    def __init__(self, blocks, _stored_block_shapes=None, check=True):

        self._stored_nb_blocks = (len(blocks), len(blocks[0]))

        self._stored_blocks = np.empty(self._stored_nb_blocks, dtype=object)
        for i in range(len(blocks)):
            for j in range(len(blocks[i])):
                self._stored_blocks[i, j] = blocks[i][j]

        if _stored_block_shapes is None:
            self._stored_block_shapes = ([block.shape[0] for block in self._stored_blocks[:, 0]],
                                         [block.shape[1] for block in self._stored_blocks[0, :]])
        else:
            # To avoid going down the tree if it is already known.
            self._stored_block_shapes = _stored_block_shapes

        # Block shape of the full matrix
        self.nb_blocks = self._compute_nb_blocks()

        # Total shape of the full matrix
        self.shape = self._compute_shape()

        LOG.debug(f"New block matrix: %s", self)

        if check:
            assert self._check_dimensions_of_blocks()
            assert self._check_dtype()

    def _compute_shape(self):
        # In a dedicated routine because it will be overloaded by subclasses.
        return sum(self._stored_block_shapes[0]), sum(self._stored_block_shapes[1])

    def _compute_nb_blocks(self):
        return self._stored_nb_blocks

    def _check_dimensions_of_blocks(self) -> bool:
        """Check that the dimensions of the blocks are consistent."""
        if not all(block.ndim == self.ndim for line in self._stored_blocks for block in line):
            return False

        for line in self.all_blocks:
            block_height = line[0].shape[0]
            for block in line[1:]:
                if not block.shape[0] == block_height:  # Same height on a given line
                    return False

        for col in np.moveaxis(self.all_blocks, 1, 0):
            block_width = col[0].shape[1]
            for block in col[1:]:
                if not block.shape[1] == block_width:  # Same width on a given column
                    return False
        return True

    def _check_dtype(self) -> bool:
        """Check that the type of the blocks are consistent."""
        for line in self._stored_blocks:
            for block in line:
                if block.dtype != self.dtype:
                    return False
        return True

    # ACCESSING DATA

    @property
    def dtype(self):
        """The type of data of all of the subblocks."""
        try:
            return self._stored_blocks[0][0].dtype
        except AttributeError:
            return None

    @property
    def all_blocks(self) -> np.ndarray:
        """The matrix of matrices. For a full block matrix, all the blocks are stored in memory."""
        return self._stored_blocks

    @property
    def block_shapes(self) -> Tuple[List[int], List[int]]:
        """The shapes of the blocks composing the BlockMatrix.

        Example::

            AAAABB
            AAAABB  ->  block_shapes = ([3], [4, 2])
            AAAABB
        """
        return self._stored_block_shapes

    def _stored_block_positions(self, global_frame=(0, 0)) -> Iterable[List[Tuple[int, int]]]:
        """The position of each blocks in the matrix as a generator.
        The list is used by subclasses where the same block may appear several times in different positions.

        Parameters
        ----------
        global_frame: Tuple[int], optional
            the coordinate of the top right corner. Default: 0, 0.

        Example::

            AAAABB
            AAAABB  ->  list(matrix._stored_block_positions) = [[(0,0)], [(0, 4)], [(2, 0)], [(2, 4)]]
            CCCCDD
        """
        x_acc = accumulate([0] + self.block_shapes[0][:-1])
        y_acc = accumulate([0] + self.block_shapes[1][:-1])
        return ([(global_frame[0] + x, global_frame[1] + y)] for x, y in product(x_acc, y_acc))

    def _put_in_full_matrix(self, full_matrix: np.ndarray, where=(0, 0)) -> None:
        """Copy the content of the block matrix in a matrix, which is modified in place."""
        all_blocks_in_flat_iterator = (block for line in self._stored_blocks for block in line)
        positions_of_all_blocks = self._stored_block_positions(global_frame=where)
        for block, positions_of_the_block in zip(all_blocks_in_flat_iterator, positions_of_all_blocks):
            if isinstance(block, BlockMatrix):
                position_of_first_appearance = positions_of_the_block[0]
                frame_of_first_appearance = (slice(position_of_first_appearance[0], position_of_first_appearance[0]+block.shape[0]),
                                             slice(position_of_first_appearance[1], position_of_first_appearance[1]+block.shape[1]))
                block._put_in_full_matrix(full_matrix, where=position_of_first_appearance)

                for position in positions_of_the_block[1:]:  # For the other appearances, only copy the first appearance
                    block_frame = (slice(position[0], position[0]+block.shape[0]),
                                   slice(position[1], position[1]+block.shape[1]))
                    full_matrix[block_frame] = full_matrix[frame_of_first_appearance]

            else:
                full_block = block if isinstance(block, np.ndarray) else block.full_matrix()
                for position in positions_of_the_block:
                    block_frame = (slice(position[0], position[0]+block.shape[0]),
                                   slice(position[1], position[1]+block.shape[1]))
                    full_matrix[block_frame] = full_block

    def full_matrix(self, dtype=None) -> np.ndarray:
        """Flatten the block structure and return a full matrix."""
        if dtype is None: dtype = self.dtype
        full_matrix = np.empty(self.shape, dtype=dtype)
        self._put_in_full_matrix(full_matrix)
        return full_matrix

    def __array__(self, dtype=None):
        return self.full_matrix(dtype=dtype)

    def no_toeplitz(self):
        """Recursively replace the block toeplitz matrices by usual block matrices.
        WARNING: the block matrices may still contain several references to the same block."""
        blocks = [[block.no_toeplitz() if isinstance(block, BlockMatrix) else block for block in line] for line in self.all_blocks]
        return BlockMatrix(blocks)

    def __deepcopy__(self, memo):
        from copy import deepcopy
        blocks = [[deepcopy(block) for block in line] for line in self._stored_blocks]
        # The call to deepcopy does not use the memo on purpose:
        # the goal is to replace references to the same block by references to different copies of the block.
        return self.__class__(blocks)

    @property
    def stored_data_size(self):
        """Return the number of entries actually stored in memory."""
        size = 0
        for line in self._stored_blocks:
            for block in line:
                if isinstance(block, np.ndarray):
                    size += np.prod(block.shape)
                else:
                    size += block.stored_data_size
        return size

    @property
    def density(self):
        return self.stored_data_size/np.prod(self.shape)

    @property
    def sparcity(self):
        return 1 - self.density

    def __hash__(self):
        # Temporary
        return id(self)

    # TRANSFORMING DATA

    def _apply_unary_op(self, op: Callable) -> 'BlockMatrix':
        """Helper function applying a function recursively on all submatrices."""
        LOG.debug(f"Apply op {op.__name__} to {self}")
        result = [[op(block) for block in line] for line in self._stored_blocks]
        return self.__class__(result, _stored_block_shapes=self._stored_block_shapes, check=False)

    def _apply_binary_op(self, op: Callable, other: 'BlockMatrix') -> 'BlockMatrix':
        """Helper function applying a binary operator recursively on all submatrices."""
        if isinstance(other, self.__class__) and self.nb_blocks == other.nb_blocks:
            LOG.debug(f"Apply op {op.__name__} to {self} and {other}")
            result = [
                [op(block, other_block) for block, other_block in zip(line, other_line)]
                for line, other_line in zip(self._stored_blocks, other._stored_blocks)
            ]
            return self.__class__(result, _stored_block_shapes=self._stored_block_shapes, check=False)
        else:
            return NotImplemented

    def __add__(self, other: 'BlockMatrix') -> 'BlockMatrix':
        from operator import add
        return self._apply_binary_op(add, other)

    def __radd__(self, other: 'BlockMatrix') -> 'BlockMatrix':
        return self + other

    def __neg__(self) -> 'BlockMatrix':
        from operator import neg
        return self._apply_unary_op(neg)

    def __sub__(self, other: 'BlockMatrix') -> 'BlockMatrix':
        from operator import sub
        return self._apply_binary_op(sub, other)

    def __rsub__(self, other: 'BlockMatrix') -> 'BlockMatrix':
        from operator import sub
        return other._apply_binary_op(sub, self)

    def __mul__(self, other: Union['BlockMatrix', Number]) -> 'BlockMatrix':
        if isinstance(other, Number):
            return self._apply_unary_op(lambda x: other*x)
        else:
            from operator import mul
            return self._apply_binary_op(mul, other)

    def __rmul__(self, other: Union['BlockMatrix', Number]) -> 'BlockMatrix':
        return self * other

    def __truediv__(self, other: Union['BlockMatrix', Number]) -> 'BlockMatrix':
        from numbers import Number
        if isinstance(other, Number):
            return self._apply_unary_op(lambda x: x/other)
        else:
            from operator import truediv
            return self._apply_binary_op(truediv, other)

    def __rtruediv__(self, other: Union['BlockMatrix', Number]) -> 'BlockMatrix':
        from numbers import Number
        if isinstance(other, Number):
            return self._apply_unary_op(lambda x: other/x)
        else:
            return self._apply_binary_op(lambda x, y: y/x, other)

    def matvec(self, other):
        """Matrix vector product.
        Named as such to be used as scipy LinearOperator."""
        LOG.debug(f"Multiplication of {self} with a full vector of size {other.shape}.")
        result = np.zeros(self.shape[0], dtype=other.dtype)
        line_heights = self.block_shapes[0]
        line_positions = list(accumulate(chain([0], line_heights)))
        col_widths = self.block_shapes[1]
        col_positions = list(accumulate(chain([0], col_widths)))
        for line, line_position, line_height in zip(self.all_blocks, line_positions, line_heights):
            line_slice = slice(line_position, line_position+line_height)
            for block, col_position, col_width in zip(line, col_positions, col_widths):
                col_slice = slice(col_position, col_position+col_width)
                result[line_slice] += block @ other[col_slice]
        return result

    def rmatvec(self, other):
        """Vector matrix product.
        Named as such to be used as scipy LinearOperator."""
        LOG.debug(f"Multiplication of a full vector of size {other.shape} with {self}.")
        result = np.zeros(self.shape[1], dtype=other.dtype)
        line_heights = self.block_shapes[0]
        line_positions = list(accumulate(chain([0], line_heights)))
        col_widths = self.block_shapes[1]
        col_positions = list(accumulate(chain([0], col_widths)))
        for col, col_position, col_width in zip(self.all_blocks.T, col_positions, col_widths):
            col_slice = slice(col_position, col_position+col_width)
            for block, line_position, line_height in zip(col, line_positions, line_heights):
                line_slice = slice(line_position, line_position+line_height)
                if isinstance(block, BlockMatrix):
                    result[col_slice] += block.rmatvec(other[line_slice])
                else:
                    result[col_slice] += other[line_slice] @ block
        return result

    def matmat(self, other):
        """Matrix-matrix product."""
        if isinstance(other, BlockMatrix) and self.block_shapes[1] == other.block_shapes[0]:
            LOG.debug(f"Multiplication of %s with %s", self, other)
            own_blocks = self.all_blocks
            other_blocks = np.moveaxis(other.all_blocks, 1, 0)
            new_matrix = []
            for own_line in own_blocks:
                new_line = []
                for other_col in other_blocks:
                    new_line.append(sum(own_block @ other_block for own_block, other_block in zip(own_line, other_col)))
                new_matrix.append(new_line)
            return BlockMatrix(new_matrix, check=False)

        elif isinstance(other, np.ndarray) and self.shape[1] == other.shape[0]:
            LOG.debug(f"Multiplication of {self} with a full matrix of shape {other.shape}.")
            # Cut the matrix and recursively call itself to use the code above.
            from capytaine.matrices.builders import cut_matrix
            cut_other = cut_matrix(other, self.block_shapes[1], [other.shape[1]], check=False)
            return (self @ cut_other).full_matrix()

    def __matmul__(self, other: Union['BlockMatrix', np.ndarray]) -> Union['BlockMatrix', np.ndarray]:
        if not (isinstance(other, BlockMatrix) or isinstance(other, np.ndarray)):
            return NotImplemented
        elif other.ndim == 2:  # Other is a matrix
            if other.shape[1] == 1:  # Actually a column vector
                return self.matvec(other.flatten())
            else:
                return self.matmat(other)
        elif other.ndim == 1:  # Other is a vector
            return self.matvec(other)
        else:
            return NotImplemented

    def __rmatmul__(self, other: Union['BlockMatrix', np.ndarray]) -> Union['BlockMatrix', np.ndarray]:
        if not (isinstance(other, BlockMatrix) or isinstance(other, np.ndarray)):
            return NotImplemented
        elif other.ndim == 2:  # Other is a matrix
            if other.shape[1] == 1:  # Actually a column vector
                return self.rmatvec(other.flatten())
            else:
                return NotImplemented
        elif other.ndim == 1:  # Other is a vector
            return self.rmatvec(other)
        else:
            return NotImplemented

    def astype(self, dtype: np.dtype) -> 'BlockMatrix':
        return self._apply_unary_op(lambda x: x.astype(dtype))

    def fft_of_list(*block_matrices, check=True):
        """Compute the fft of a list of block matrices of the same type and shape.
        The output is a list of block matrices of the same shape as the input ones.
        The fft is computed element-wise, so the block structure does not cause any mathematical difficulty.
        Returns an array of BlockMatrices.
        """
        class_of_matrices = type(block_matrices[0])
        nb_blocks = block_matrices[0]._stored_nb_blocks

        LOG.debug(f"FFT of {len(block_matrices)} {class_of_matrices.__name__} (stored blocks = {nb_blocks})")

        if check:
            # Check the validity of the shapes of the matrices given as input
            shape = block_matrices[0].shape
            assert [nb_blocks == matrix._stored_nb_blocks for matrix in block_matrices[1:]]
            assert [shape == matrix.shape for matrix in block_matrices[1:]]
            assert [class_of_matrices == type(matrix) for matrix in block_matrices[1:]]

        # Initialize a vector of block matrices without values in the blocks.
        result = np.empty(len(block_matrices), dtype=object)
        for i in range(len(block_matrices)):
            result[i] = class_of_matrices(np.empty(nb_blocks, dtype=object),
                                          _stored_block_shapes=block_matrices[0]._stored_block_shapes,
                                          check=False)

        for i_block, j_block in product(range(nb_blocks[0]), range(nb_blocks[1])):
            list_of_i_j_blocks = [block_matrices[i_matrix]._stored_blocks[i_block, j_block]
                                  for i_matrix in range(len(block_matrices))]

            if any(isinstance(block, np.ndarray) or isinstance(block, LowRankMatrix) for block in list_of_i_j_blocks):
                list_of_i_j_blocks = [block if isinstance(block, np.ndarray) else block.full_matrix() for block in list_of_i_j_blocks]
                fft_of_blocks = np.fft.fft(list_of_i_j_blocks, axis=0)
            else:
                fft_of_blocks = BlockMatrix.fft_of_list(*list_of_i_j_blocks, check=False)

            for matrix, computed_block in zip(result, fft_of_blocks):
                matrix._stored_blocks[i_block, j_block] = computed_block

        return result

    # COMPARISON AND REDUCTION

    def __eq__(self, other: 'BlockMatrix') -> 'BlockMatrix[bool]':
        from operator import eq
        return self._apply_binary_op(eq, other)

    def __invert__(self) -> 'BlockMatrix':
        """Boolean not (~)"""
        from operator import invert
        return self._apply_unary_op(invert)

    def __ne__(self, other: 'BlockMatrix') -> 'BlockMatrix[bool]':
        return ~(self == other)

    def all(self) -> bool:
        for line in self._stored_blocks:
            for block in line:
                if not block.all():
                    return False
        return True

    def any(self) -> bool:
        for line in self._stored_blocks:
            for block in line:
                if block.any():
                    return True
        return False

    def min(self) -> Number:
        return min(block.min() for line in self._stored_blocks for block in line)

    def max(self) -> Number:
        return max(block.max() for line in self._stored_blocks for block in line)

    # DISPLAYING DATA

    @property
    def str_shape(self):
        blocks_str = []
        for line in self.all_blocks:
            for block in line:
                if isinstance(block, BlockMatrix):
                    blocks_str.append(block.str_shape)
                elif isinstance(block, np.ndarray) or isinstance(block, LowRankMatrix):
                    blocks_str.append("{}×{}".format(*block.shape))
                else:
                    blocks_str.append("?×?")

        if len(set(blocks_str)) == 1:
            return "{}×{}×[".format(*self.nb_blocks) + blocks_str[0] + "]"
        else:
            blocks_str = np.array(blocks_str).reshape(self.nb_blocks).tolist()
            return str(blocks_str).replace("'", "")

    def __str__(self):
        if not hasattr(self, '_str'):
            args = [self.str_shape]
            if self.dtype not in [np.float64, float]:
                args.append(f"dtype={self.dtype}")
            self._str = f"{self.__class__.__name__}(" + ", ".join(args) + ")"
        return self._str

    def _repr_pretty_(self, p, cycle):
        p.text(self.__str__())

    display_color = cycle([f'C{i}' for i in range(10)])

    def _patches(self,
                 global_frame: Union[Tuple[int, int], np.ndarray]
                 ):
        """Helper function for displaying the shape of the matrix.
        Recursively returns a list of rectangles representing the sub-blocks of the matrix.

        Uses BlockMatrix.display_color to assign color to the blocks.
        By default, it cycles through matplotlib default colors.
        But if display_color is redefined as a callable, it is called with the block as argument.

        Parameters
        ----------
        global_frame: tuple of ints
            coordinates of the origin in the top left corner.

        Returns
        -------
        list of matplotlib.patches.Rectangle
        """
        matplotlib_patches = import_optional_dependency("matplotlib.patches", "matplotlib")
        Rectangle = matplotlib_patches.Rectangle

        all_blocks_in_flat_iterator = (block for line in self._stored_blocks for block in line)
        positions_of_all_blocks = self._stored_block_positions(global_frame=global_frame)
        patches = []
        for block, positions_of_the_block in zip(all_blocks_in_flat_iterator, positions_of_all_blocks):
            position_of_first_appearance = positions_of_the_block[0]
            # Exchange coordinates: row index i -> y, column index j -> x
            position_of_first_appearance = np.array((position_of_first_appearance[1], position_of_first_appearance[0]))

            if isinstance(block, BlockMatrix):
                patches_of_this_block = block._patches(np.array((position_of_first_appearance[1], position_of_first_appearance[0])))
            elif isinstance(block, np.ndarray):

                if isinstance(self.display_color, Iterator):
                    color = next(self.display_color)
                elif callable(self.display_color):
                    color = self.display_color(block)
                else:
                    color = np.random.rand(3)

                patches_of_this_block = [Rectangle(position_of_first_appearance,
                                                   block.shape[1], block.shape[0],
                                                   edgecolor='k', facecolor=color)]
            elif isinstance(block, LowRankMatrix):

                if isinstance(self.display_color, Iterator):
                    color = next(self.display_color)
                elif callable(self.display_color):
                    color = self.display_color(block)
                else:
                    color = np.random.rand(3)

                patches_of_this_block = [
                    # Left block
                    Rectangle(position_of_first_appearance,
                              block.left_matrix.shape[1], block.left_matrix.shape[0],
                              edgecolor='k', facecolor=color),
                    # Top block
                    Rectangle(position_of_first_appearance,
                              block.right_matrix.shape[1], block.right_matrix.shape[0],
                              edgecolor='k', facecolor=color),
                    # Rest of the matrix
                    Rectangle(position_of_first_appearance,
                              block.right_matrix.shape[1], block.left_matrix.shape[0],
                              facecolor=color, alpha=0.2),
                ]
            else:
                raise NotImplementedError()

            patches.extend(patches_of_this_block)

            # For the other appearances, copy the patches of the first appearance
            for block_position in positions_of_the_block[1:]:
                block_position = np.array((block_position[1], block_position[0]))
                for patch in patches_of_this_block:  # A block can be made of several patches.
                    shift = block_position - position_of_first_appearance
                    patch_position = np.array(patch.get_xy()) + shift
                    patches.append(Rectangle(patch_position, patch.get_width(), patch.get_height(),
                                             facecolor=patch.get_facecolor(), alpha=0.2))

        return patches

    def plot_shape(self):
        """Plot the structure of the matrix using matplotlib."""
        matplotlib = import_optional_dependency("matplotlib")
        plt = matplotlib.pyplot

        plt.figure()
        for patch in self._patches((0, 0)):
            plt.gca().add_patch(patch)
        plt.axis('equal')
        plt.xlim(0, self.shape[1])
        plt.ylim(0, self.shape[0])
        plt.gca().invert_yaxis()
        # plt.show()


    def access_block_by_path(self, path):
        """
        Access a diagonal block in a block matrix from the path of the
        corresponding leaf
        """
        this_block = self
        for index in path:
            this_block = this_block.all_blocks[index, index]
        return this_block


#!/usr/bin/env python

import numpy as np
import xarray as xr
import capytaine as cpt
import matplotlib.pyplot as plt
from scipy.optimize import newton

# We consider a vertical cylinder clamped at the sea bottom and freely moving at the top above the free surface.

water_depth = 10

cylinder_length = 12.0
cylinder_density = 1000.0
cylinder_diameter = 1.0
young_modulus = 1e9
inertia_moment = np.pi * cylinder_diameter**4 / 64  # Cylinder with circular section

# The following code computes the shapes of the deformation modes of the body and the associated eigenfrequencies.

def deformation_mode(i_mode, z):
    alpha = newton(lambda a: np.cos(a)*np.cosh(a) + 1, (2*i_mode + 1)*np.pi/2)
    coeff = (-np.cos(alpha) - np.cosh(alpha))/(np.sin(alpha) + np.sinh(alpha))
    k = alpha/cylinder_length
    z_ = z + water_depth
    return 0.5 * (np.cos(k*z_) - np.cosh(k*z_) + coeff*(np.sin(k*z_) - np.sinh(k*z_)))

def eigenfrequency(i_mode):
    alpha = newton(lambda a: np.cos(a)*np.cosh(a) + 1, (2*i_mode + 1)*np.pi/2)
    return alpha/cylinder_length**2 * np.sqrt(young_modulus*inertia_moment/(cylinder_density*np.pi*(cylinder_diameter/2)**2))


# Plotting the x-deformation of the first 4 modes
z_range = np.linspace(-water_depth, -water_depth + cylinder_length, 100)

# for i_mode in range(4):
#     plt.plot(deformation_mode(i_mode, z_range), z_range, label=f"mode_{i_mode}")
# plt.legend()


mesh = cpt.mesh_vertical_cylinder(
        center=(0.0, 0.0, cylinder_length/2 - water_depth - 0.1),
        length=cylinder_length,
        radius=cylinder_diameter/2,
        resolution=(4, 15, 36),
        )

nb_modes = 10
dofs = {}
for i in range(nb_modes):
    dofs[f"mode_{i}"] = np.array([(deformation_mode(i, z), 0.0, 0.0) for x, y, z in mesh.faces_centers])

full_body = cpt.FloatingBody(mesh=mesh, dofs=dofs, center_of_mass=(0, 0, cylinder_length/2 - water_depth))
body = full_body.immersed_part(water_depth=water_depth)

body.inertia_matrix = body.add_dofs_labels_to_matrix(np.eye(nb_modes))
body.internal_stiffness_matrix = body.add_dofs_labels_to_matrix(np.diag([eigenfrequency(i_mode) for i_mode in range(nb_modes)]))
body.hydrostatic_stiffness = body.compute_hydrostatic_stiffness()

test_matrix = xr.Dataset(coords={
    "omega": [1.0],
    "radiating_dof": list(body.dofs),
    "wave_direction": [0.0],
    "water_depth": [water_depth],
    })
ds = cpt.BEMSolver().fill_dataset(test_matrix, body)
rao = cpt.post_pro.rao(ds, stiffness=body.internal_stiffness_matrix)


deformation_profile = sum(
    rao.isel(radiating_dof=i_mode).values[0] * deformation_mode(i_mode, z_range)
    for i_mode, dof in enumerate(body.dofs)
    )
plt.plot(np.real(deformation_profile), z_range)

plt.show()




import pytest

import numpy as np

from capytaine.meshes.meshes import Mesh
from capytaine.meshes.collections import CollectionOfMeshes
from capytaine.meshes.symmetric import TranslationalSymmetricMesh, ReflectionSymmetricMesh, AxialSymmetricMesh

from capytaine.bodies.predefined.rectangles import Rectangle, OpenRectangularParallelepiped, RectangularParallelepiped
from capytaine.bodies.predefined.spheres import Sphere
from capytaine.bodies.predefined.cylinders import Disk, HorizontalCylinder, VerticalCylinder


def test_rectangle_generation():
    # Rejects invalid inputs:
    with pytest.raises(AssertionError):
        Rectangle(size=(5.0, 6.0, 7.0))

    with pytest.raises(AssertionError):
        Rectangle(size=(-4, 5))

    with pytest.raises(AssertionError):
        Rectangle(resolution=(4.1, 5))

    with pytest.raises(ValueError):
        Rectangle(resolution=(3, 3), reflection_symmetry=True)

    # Test mesh with symmetry
    rec = Rectangle(size=(10, 10), resolution=(6, 2), normal=(1, 0, 0),
                    reflection_symmetry=False, translational_symmetry=False,
                    center=(0, 0, -5), name="test")

    assert rec.name == "test"
    assert isinstance(rec.mesh, Mesh)
    assert rec.mesh.name == "test_mesh"
    assert all(rec.size == [10.0, 10.0])
    assert all(rec.geometric_center == [0, 0, -5.0])
    assert rec.mesh.nb_faces == 12
    assert rec.mesh.nb_vertices == 21

    # x coordinate
    assert np.allclose(rec.mesh.vertices[:, 0], 0.0)

    # y coordinate
    assert np.all(rec.mesh.vertices[:, 1] <= 5.0)
    assert np.all(rec.mesh.vertices[:, 1] >= -5.0)

    # z coordinate
    assert np.all(rec.mesh.vertices[:, 2] <= 0.0)
    assert np.all(rec.mesh.vertices[:, 2] >= -10.0)

    # Test mesh with reflection symmetry
    sym_rec = Rectangle(size=(10, 10), resolution=(6, 2),
                        translational_symmetry=False, reflection_symmetry=True,
                        center=(0, 0, -5))
    assert isinstance(sym_rec.mesh, ReflectionSymmetricMesh)
    assert sym_rec.mesh.nb_submeshes == 2
    assert sym_rec.mesh.nb_faces == 12

    sym_rec_merged_mesh = sym_rec.mesh.merged()
    assert isinstance(sym_rec_merged_mesh, Mesh)
    # assert sym_rec_merged_mesh == rec.mesh

    # Test mesh with translation symmetry
    trans_rec = Rectangle(size=(10, 10), resolution=(6, 2),
                          translational_symmetry=True, reflection_symmetry=False,
                          center=(0, 0, -5))
    assert isinstance(trans_rec.mesh, TranslationalSymmetricMesh)
    assert trans_rec.mesh.nb_submeshes == 6
    assert trans_rec.mesh.nb_faces == 12

    trans_rec_merged_mesh = trans_rec.mesh.merged()
    assert isinstance(trans_rec_merged_mesh, Mesh)
    # assert trans_rec_merged_mesh == rec.mesh


def test_open_parallelepiped_generation():
    para = OpenRectangularParallelepiped(size=(5.0, 1.0, 3.0), center=(0, 0, -1.5),
                                         resolution=(5, 2, 3),
                                         translational_symmetry=False, name="test")
    assert para.name == "test"
    assert isinstance(para.mesh, Mesh)
    assert para.mesh.name == "test_mesh"
    assert para.mesh.nb_faces == 42

    assert np.isclose(np.abs(para.mesh.vertices[:, 0]).max(), 2.5)
    assert np.isclose(np.abs(para.mesh.vertices[:, 1]).max(), 0.5)
    assert np.all(para.mesh.vertices[:, 2] <= 0.0)
    assert np.all(para.mesh.vertices[:, 2] >= -3.0)

    clever_para = OpenRectangularParallelepiped(size=(5.0, 1.0, 3.0), center=(0, 0, -1.5),
                                                resolution=(5, 2, 3),
                                                translational_symmetry=True, name="clever_test")

    assert clever_para.mesh.nb_faces == 42
    assert isinstance(clever_para.mesh, CollectionOfMeshes)
    assert any(isinstance(submesh, TranslationalSymmetricMesh) for submesh in clever_para.mesh)


def test_parallelepiped_generation():
    full_para = RectangularParallelepiped(size=(5.0, 1.0, 3.0), center=(0, 0, -1.5),
                                          resolution=(6, 2, 3),
                                          translational_symmetry=False, name="full_test")
    assert full_para.mesh.nb_faces == 72
    # full_para.show()

    sym_para = RectangularParallelepiped(size=(5.0, 1.0, 3.0), center=(0, 0, -1.5),
                                         resolution=(6, 2, 3),
                                         reflection_symmetry=True, name="sym_full_test")
    assert sym_para.mesh.nb_faces == 72
    # sym_para.show()

    trans_para = RectangularParallelepiped(size=(5.0, 1.0, 3.0), center=(0, 0, -1.5),
                                           resolution=(6, 2, 3),
                                           translational_symmetry=True, name="trans_full_test")
    assert trans_para.mesh.nb_faces == 72
    # trans_para.show()


def test_disk():
    a = Disk(resolution=(6, 6))
    b = Disk(resolution=(6, 6), axial_symmetry=True)
    # print(b.mesh.tree_view())
    c = Disk(resolution=(6, 6), reflection_symmetry=True)
    # TODO

def test_cylinder():
    HorizontalCylinder()
    VerticalCylinder()
    # TODO


def test_cylinder_submesh_names():
    for s in [True, False]:
        c = HorizontalCylinder(reflection_symmetry=True, translation_symmetry=s, name="cylinder")
        assert c.mesh[0].name == "half_cylinder_mesh"
        assert c.mesh[1].name == "mirrored_of_half_cylinder_mesh"


############
#  SPHERE  #
############

def test_sphere_name():
    sphere = Sphere()
    assert sphere.name.startswith("sphere_")

def test_sphere_axisymmetric():
    sphere = Sphere(axial_symmetry=True)
    assert isinstance(sphere.mesh, AxialSymmetricMesh)

def test_sphere_not_axisymmetric():
    sphere = Sphere(axial_symmetry=False)
    assert isinstance(sphere.mesh, Mesh)

def test_sphere_nb_panels():
    sphere = Sphere(ntheta=5, nphi=5, clip_free_surface=False)
    assert sphere.mesh.nb_faces == 25

def test_sphere_nb_panels_clipped():
    sphere = Sphere(ntheta=5, nphi=5, clip_free_surface=True)
    assert sphere.mesh.nb_faces == 25

def test_sphere_nb_panels_clipped_is_underwater():
    sphere = Sphere(ntheta=5, nphi=5, clip_free_surface=True)
    assert np.all(sphere.mesh.vertices[:, 2] <= 0.0)

def test_sphere_out_of_water():
    with pytest.raises(ValueError):
        sphere = Sphere(radius=1.0, center=(0, 0, 10), clip_free_surface=True)

def test_sphere_geometric_center():
    sphere = Sphere(radius=2.0, center=(-2.0, 2.0, 1.0))
    assert np.allclose(sphere.geometric_center, np.array([-2.0, 2.0, 1.0]))

def test_sphere_clipping():
    s1 = Sphere(ntheta=4, nphi=4, axial_symmetry=False, clip_free_surface=True)
    s2 = Sphere(ntheta=8, nphi=4, axial_symmetry=False, clip_free_surface=False).keep_immersed_part()
    assert s1.mesh.nb_faces == s2.mesh.nb_faces
    assert s1.mesh.nb_vertices == s2.mesh.nb_vertices
    # TODO: test that the faces are actually the same.



#!/usr/bin/env python

import numpy as np
import capytaine as cpt

# Generate the mesh of a cylinder
cylinder = cpt.FloatingBody(
        mesh=cpt.mesh_horizontal_cylinder(
            length=10.0, radius=1.0,
            center=(0, 0, -2),
            resolution=(1, 6, 8)
            ))

engine = cpt.BasicMatrixEngine()
green_function = cpt.Delhommeau()

S, K = engine.build_matrices(
    cylinder.mesh, cylinder.mesh,
    free_surface=0.0, water_depth=np.inf,
    wavenumber=1.0,
    green_function=green_function,
)

# Plot the absolute value of the matrix S
import matplotlib.pyplot as plt
plt.imshow(abs(np.array(S)))
plt.colorbar()
plt.title("$|S|$")
plt.tight_layout()
plt.show()



import numpy as np
import capytaine as cpt

vertices = np.loadtxt("vertices.dat")
normals = np.loadtxt("normal.dat")
mesh = cpt.Mesh(vertices, np.arange(0, vertices.shape[0]).reshape(-1, 4))
mesh.show()



import numpy as np
from numpy import pi
import xarray as xr
import matplotlib.pyplot as plt
import capytaine as cpt

cpt.set_logging('INFO')

def make_cylinder(resolution):
    """Make cylinder with a mesh of a given resolution in panels/meter."""
    radius = 1.0
    length = 5.0
    mesh = cpt.mesh_horizontal_cylinder(
        length=length, radius=radius,
        center=(0, 0, -1.5*radius),
        resolution=(int(resolution*radius),
                    int(2*pi*length*resolution),
                    int(length*resolution))
    )
    body = cpt.FloatingBody(mesh, name=f"cylinder_{mesh.nb_faces:04d}")
    # Store the number of panels in the name of the body
    body.add_translation_dof(name="Heave")
    return body

test_matrix = xr.Dataset(coords={
    'omega': [1.0],
    'radiating_dof': ['Heave'],
})

bodies = [make_cylinder(n) for n in np.linspace(1, 5, 10)]
ds1 = cpt.BEMSolver(green_function=cpt.XieDelhommeau()).fill_dataset(test_matrix, bodies)

def read_nb_faces_in_mesh_name(ds):
    """Read the name of the body to guess the resolution of the mesh."""
    ds.coords['nb_faces'] = xr.DataArray([int(name[9:]) for name in ds['body_name'].values], coords=[ds['body_name']])
    ds = ds.swap_dims({'body_name': 'nb_faces'})
    return ds
ds1 = read_nb_faces_in_mesh_name(ds1)

ds1['added_mass'].plot(x='nb_faces')

plt.show()



# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>
"""Solver for the BEM problem.

.. code-block:: python

    problem = RadiationProblem(...)
    result = BEMSolver(green_functions=..., engine=...).solve(problem)

"""

import logging

import numpy as np

from datetime import datetime

from rich.progress import track

from capytaine.bem.problems_and_results import LinearPotentialFlowProblem
from capytaine.green_functions.delhommeau import Delhommeau
from capytaine.bem.engines import BasicMatrixEngine
from capytaine.io.xarray import problems_from_dataset, assemble_dataset, kochin_data_array
from capytaine.tools.optional_imports import silently_import_optional_dependency
from capytaine.tools.lists_of_points import _normalize_points, _normalize_free_surface_points
from capytaine.tools.symbolic_multiplication import supporting_symbolic_multiplication

LOG = logging.getLogger(__name__)

class BEMSolver:
    """
    Solver for linear potential flow problems.

    Parameters
    ----------
    green_function: AbstractGreenFunction, optional
        Object handling the computation of the Green function.
        (default: :class:`~capytaine.green_function.delhommeau.Delhommeau`)
    engine: MatrixEngine, optional
        Object handling the building of matrices and the resolution of linear systems with these matrices.
        (default: :class:`~capytaine.bem.engines.BasicMatrixEngine`)

    Attributes
    ----------
    exportable_settings : dict
        Settings of the solver that can be saved to reinit the same solver later.
    """

    def __init__(self, *, green_function=None, engine=None):
        self.green_function = Delhommeau() if green_function is None else green_function
        self.engine = BasicMatrixEngine() if engine is None else engine

        try:
            self.exportable_settings = {
                **self.green_function.exportable_settings,
                **self.engine.exportable_settings
            }
        except AttributeError:
            pass

    def __str__(self):
        return f"BEMSolver(engine={self.engine}, green_function={self.green_function})"

    def __repr__(self):
        return self.__str__()

    def _repr_pretty_(self, p, cycle):
        p.text(self.__str__())

    @classmethod
    def from_exported_settings(settings):
        raise NotImplementedError

    def solve(self, problem, method='indirect', keep_details=True, _check_wavelength=True):
        """Solve the linear potential flow problem.

        Parameters
        ----------
        problem: LinearPotentialFlowProblem
            the problem to be solved
        method: string, optional
            select boundary integral approach indirect (i.e.Nemoh)/direct (i.e.WAMIT) (default: indirect)
        keep_details: bool, optional
            if True, store the sources and the potential on the floating body in the output object
            (default: True)
        _check_wavelength: bool, optional
            if True, check the mesh resolution with respect to the wavelength

        Returns
        -------
        LinearPotentialFlowResult
            an object storing the problem data and its results
        """
        LOG.info("Solve %s.", problem)

        if _check_wavelength: self._check_wavelength([problem])

        if problem.forward_speed != 0.0:
            omega, wavenumber = problem.encounter_omega, problem.encounter_wavenumber
        else:
            omega, wavenumber = problem.omega, problem.wavenumber

        linear_solver = supporting_symbolic_multiplication(self.engine.linear_solver)
        if (method == 'direct'):
            if problem.forward_speed != 0.0:
                raise NotImplementedError("Direct solver is not able to solve problems with forward speed.")

            S, D = self.engine.build_matrices(
                    problem.body.mesh, problem.body.mesh,
                    problem.free_surface, problem.water_depth, wavenumber,
                    self.green_function, adjoint_double_layer=False
                    )

            potential = linear_solver(D, S @ problem.boundary_condition)
            pressure = 1j * omega * problem.rho * potential
            sources = None
        else:
            S, K = self.engine.build_matrices(
                    problem.body.mesh, problem.body.mesh,
                    problem.free_surface, problem.water_depth, wavenumber,
                    self.green_function, adjoint_double_layer=True
                    )

            sources = linear_solver(K, problem.boundary_condition)
            potential = S @ sources
            pressure = 1j * omega * problem.rho * potential
            if problem.forward_speed != 0.0:
                result = problem.make_results_container(sources=sources)
                # Temporary result object to compute the ∇Φ term
                nabla_phi = self._compute_potential_gradient(problem.body.mesh, result)
                pressure += problem.rho * problem.forward_speed * nabla_phi[:, 0]

        forces = problem.body.integrate_pressure(pressure)

        if not keep_details:
            result = problem.make_results_container(forces)
        else:
            result = problem.make_results_container(forces, sources, potential, pressure)

        LOG.debug("Done!")

        return result

    def solve_all(self, problems, *, method='indirect', n_jobs=1, progress_bar=True, _check_wavelength=True, **kwargs):
        """Solve several problems.
        Optional keyword arguments are passed to `BEMSolver.solve`.

        Parameters
        ----------
        problems: list of LinearPotentialFlowProblem
            several problems to be solved
        method: string, optional
            select boundary integral approach indirect (i.e.Nemoh)/direct (i.e.WAMIT) (default: indirect)
        n_jobs: int, optional (default: 1)
            the number of jobs to run in parallel using the optional dependency `joblib`
            By defaults: do not use joblib and solve sequentially.
        progress_bar: bool, optional (default: True)
            Display a progress bar while solving

        Returns
        -------
        list of LinearPotentialFlowResult
            the solved problems
        """
        if _check_wavelength: self._check_wavelength(problems)

        if n_jobs == 1:  # force sequential resolution
            problems = sorted(problems)
            if progress_bar:
                problems = track(problems, total=len(problems), description="Solving BEM problems")
            return [self.solve(pb, method=method, _check_wavelength=False, **kwargs) for pb in problems]
        else:
            joblib = silently_import_optional_dependency("joblib")
            if joblib is None:
                raise ImportError(f"Setting the `n_jobs` argument to {n_jobs} requires the missing optional dependency 'joblib'.")
            groups_of_problems = LinearPotentialFlowProblem._group_for_parallel_resolution(problems)
            parallel = joblib.Parallel(return_as="generator", n_jobs=n_jobs)
            groups_of_results = parallel(joblib.delayed(self.solve_all)(grp, method=method, n_jobs=1, progress_bar=False, _check_wavelength=False, **kwargs) for grp in groups_of_problems)
            if progress_bar:
                groups_of_results = track(groups_of_results,
                                          total=len(groups_of_problems),
                                          description=f"Solving BEM problems with {n_jobs} threads:")
            results = [res for grp in groups_of_results for res in grp]  # flatten the nested list
            return results

    @staticmethod
    def _check_wavelength(problems):
        """Display a warning if some of the problems have a mesh resolution
        that might not be sufficient for the given wavelength."""
        risky_problems = [pb for pb in problems
                          if pb.wavelength < pb.body.minimal_computable_wavelength]
        nb_risky_problems = len(risky_problems)
        if nb_risky_problems == 1:
            pb = risky_problems[0]
            freq_type = risky_problems[0].provided_freq_type
            freq = pb.__getattribute__(freq_type)
            LOG.warning(f"Mesh resolution for {pb}:\n"
                    f"The resolution of the mesh of the body {pb.body.__short_str__()} might "
                    f"be insufficient for {freq_type}={freq}.\n"
                     "This warning appears because the largest panel of this mesh "
                    f"has radius {pb.body.mesh.faces_radiuses.max():.3f} > wavelength/8."
                    )
        elif nb_risky_problems > 1:
            freq_type = risky_problems[0].provided_freq_type
            freqs = np.array([float(pb.__getattribute__(freq_type)) for pb in risky_problems])
            LOG.warning(f"Mesh resolution for {nb_risky_problems} problems:\n"
                         "The resolution of the mesh might be insufficient "
                        f"for {freq_type} ranging from {freqs.min():.3f} to {freqs.max():.3f}.\n"
                         "This warning appears when the largest panel of this mesh "
                         "has radius > wavelength/8."
                    )

    def fill_dataset(self, dataset, bodies, *, method='indirect', n_jobs=1, **kwargs):
        """Solve a set of problems defined by the coordinates of an xarray dataset.

        Parameters
        ----------
        dataset : xarray Dataset
            dataset containing the problems parameters: frequency, radiating_dof, water_depth, ...
        bodies : FloatingBody or list of FloatingBody
            The body or bodies involved in the problems
            They should all have different names.
        method: string, optional
            select boundary integral approach indirect (i.e.Nemoh)/direct (i.e.WAMIT) (default: indirect)
        n_jobs: int, optional (default: 1)
            the number of jobs to run in parallel using the optional dependency `joblib`
            By defaults: do not use joblib and solve sequentially.
        progress_bar: bool, optional (default: True)
            Display a progress bar while solving

        Returns
        -------
        xarray Dataset
        """
        attrs = {'start_of_computation': datetime.now().isoformat(),
                 **self.exportable_settings}
        problems = problems_from_dataset(dataset, bodies)
        if 'theta' in dataset.coords:
            results = self.solve_all(problems, keep_details=True, method=method, n_jobs=n_jobs)
            kochin = kochin_data_array(results, dataset.coords['theta'])
            dataset = assemble_dataset(results, attrs=attrs, **kwargs)
            dataset.update(kochin)
        else:
            results = self.solve_all(problems, keep_details=False, method=method, n_jobs=n_jobs)
            dataset = assemble_dataset(results, attrs=attrs, **kwargs)
        return dataset


    def compute_potential(self, points, result):
        """Compute the value of the potential at given points for a previously solved potential flow problem.

        Parameters
        ----------
        points: array of shape (3,) or (N, 3), or 3-ple of arrays returned by meshgrid, or cpt.Mesh or cpt.CollectionOfMeshes object
            Coordinates of the point(s) at which the potential should be computed
        result: LinearPotentialFlowResult
            The return of the BEM solver

        Returns
        -------
        complex-valued array of shape (1,) or (N,) or (nx, ny, nz) or (mesh.nb_faces,) depending of the kind of input
            The value of the potential at the points

        Raises
        ------
        Exception: if the :code:`LinearPotentialFlowResult` object given as input does not contain the source distribution.
        """
        points, output_shape = _normalize_points(points, keep_mesh=True)
        if result.sources is None:
            raise Exception(f"""The values of the sources of {result} cannot been found.
            They probably have not been stored by the solver because the option keep_details=True have not been set or the direct method has been used.
            Please re-run the resolution with the indirect method and keep_details=True.""")

        S, _ = self.green_function.evaluate(points, result.body.mesh, result.free_surface, result.water_depth, result.encounter_wavenumber)
        potential = S @ result.sources  # Sum the contributions of all panels in the mesh
        return potential.reshape(output_shape)

    def _compute_potential_gradient(self, points, result):
        points, output_shape = _normalize_points(points, keep_mesh=True)

        if result.sources is None:
            raise Exception(f"""The values of the sources of {result} cannot been found.
            They probably have not been stored by the solver because the option keep_details=True have not been set.
            Please re-run the resolution with this option.""")

        _, gradG = self.green_function.evaluate(points, result.body.mesh, result.free_surface, result.water_depth, result.encounter_wavenumber,
                                                early_dot_product=False)
        velocities = np.einsum('ijk,j->ik', gradG, result.sources)  # Sum the contributions of all panels in the mesh
        return velocities.reshape((*output_shape, 3))

    def compute_velocity(self, points, result):
        """Compute the value of the velocity vector at given points for a previously solved potential flow problem.

        Parameters
        ----------
        points: array of shape (3,) or (N, 3), or 3-ple of arrays returned by meshgrid, or cpt.Mesh or cpt.CollectionOfMeshes object
            Coordinates of the point(s) at which the velocity should be computed
        result: LinearPotentialFlowResult
            The return of the BEM solver

        Returns
        -------
        complex-valued array of shape (3,) or (N,, 3) or (nx, ny, nz, 3) or (mesh.nb_faces, 3) depending of the kind of input
            The value of the velocity at the points

        Raises
        ------
        Exception: if the :code:`LinearPotentialFlowResult` object given as input does not contain the source distribution.
        """
        nabla_phi = self._compute_potential_gradient(points, result)
        if result.forward_speed != 0.0:
            nabla_phi[..., 0] -= result.forward_speed
        return nabla_phi

    def compute_pressure(self, points, result):
        """Compute the value of the pressure at given points for a previously solved potential flow problem.

        Parameters
        ----------
        points: array of shape (3,) or (N, 3), or 3-ple of arrays returned by meshgrid, or cpt.Mesh or cpt.CollectionOfMeshes object
            Coordinates of the point(s) at which the pressure should be computed
        result: LinearPotentialFlowResult
            The return of the BEM solver

        Returns
        -------
        complex-valued array of shape (1,) or (N,) or (nx, ny, nz) or (mesh.nb_faces,) depending of the kind of input
            The value of the pressure at the points

        Raises
        ------
        Exception: if the :code:`LinearPotentialFlowResult` object given as input does not contain the source distribution.
        """
        if result.forward_speed != 0:
            pressure = 1j * result.encounter_omega * result.rho * self.compute_potential(points, result)
            nabla_phi = self._compute_potential_gradient(points, result)
            pressure += result.rho * result.forward_speed * nabla_phi[..., 0]
        else:
            pressure = 1j * result.omega * result.rho * self.compute_potential(points, result)
        return pressure


    def compute_free_surface_elevation(self, points, result):
        """Compute the value of the free surface elevation at given points for a previously solved potential flow problem.

        Parameters
        ----------
        points: array of shape (2,) or (N, 2), or 2-ple of arrays returned by meshgrid, or cpt.Mesh or cpt.CollectionOfMeshes object
            Coordinates of the point(s) at which the free surface elevation should be computed
        result: LinearPotentialFlowResult
            The return of the BEM solver

        Returns
        -------
        complex-valued array of shape (1,) or (N,) or (nx, ny, nz) or (mesh.nb_faces,) depending of the kind of input
            The value of the free surface elevation at the points

        Raises
        ------
        Exception: if the :code:`LinearPotentialFlowResult` object given as input does not contain the source distribution.
        """
        points, output_shape = _normalize_free_surface_points(points, keep_mesh=True)

        if result.forward_speed != 0:
            fs_elevation = -1/result.g * (-1j*result.encounter_omega) * self.compute_potential(points, result)
            nabla_phi = self._compute_potential_gradient(points, result)
            fs_elevation += -1/result.g * result.forward_speed * nabla_phi[..., 0]
        else:
            fs_elevation = -1/result.g * (-1j*result.omega) * self.compute_potential(points, result)

        return fs_elevation.reshape(output_shape)


    ## Legacy

    def get_potential_on_mesh(self, result, mesh, chunk_size=50):
        """Compute the potential on a mesh for the potential field of a previously solved problem.
        Since the interaction matrix does not need to be computed in full to compute the matrix-vector product,
        only a few lines are evaluated at a time to reduce the memory cost of the operation.

        The newer method :code:`compute_potential` should be preferred in the future.

        Parameters
        ----------
        result : LinearPotentialFlowResult
            the return of the BEM solver
        mesh : Mesh or CollectionOfMeshes
            a mesh
        chunk_size: int, optional
            Number of lines to compute in the matrix.
            (legacy, should be passed as an engine setting instead).

        Returns
        -------
        array of shape (mesh.nb_faces,)
            potential on the faces of the mesh

        Raises
        ------
        Exception: if the :code:`Result` object given as input does not contain the source distribution.
        """
        LOG.info(f"Compute potential on {mesh.name} for {result}.")

        if result.sources is None:
            raise Exception(f"""The values of the sources of {result} cannot been found.
            They probably have not been stored by the solver because the option keep_details=True have not been set or the direct method has been used.
            Please re-run the resolution with the indirect method and keep_details=True.""")

        if chunk_size > mesh.nb_faces:
            S = self.engine.build_S_matrix(
                mesh,
                result.body.mesh,
                result.free_surface, result.water_depth, result.wavenumber,
                self.green_function
            )
            phi = S @ result.sources

        else:
            phi = np.empty((mesh.nb_faces,), dtype=np.complex128)
            for i in range(0, mesh.nb_faces, chunk_size):
                faces_to_extract = list(range(i, min(i+chunk_size, mesh.nb_faces)))
                S = self.engine.build_S_matrix(
                    mesh.extract_faces(faces_to_extract),
                    result.body.mesh,
                    result.free_surface, result.water_depth, result.wavenumber,
                    self.green_function
                )
                phi[i:i+chunk_size] = S @ result.sources

        LOG.debug(f"Done computing potential on {mesh.name} for {result}.")

        return phi

    def get_free_surface_elevation(self, result, free_surface, keep_details=False):
        """Compute the elevation of the free surface on a mesh for a previously solved problem.

        The newer method :code:`compute_free_surface_elevation` should be preferred in the future.

        Parameters
        ----------
        result : LinearPotentialFlowResult
            the return of the solver
        free_surface : FreeSurface
            a meshed free surface
        keep_details : bool, optional
            if True, keep the free surface elevation in the LinearPotentialFlowResult (default:False)

        Returns
        -------
        array of shape (free_surface.nb_faces,)
            the free surface elevation on each faces of the meshed free surface

        Raises
        ------
        Exception: if the :code:`Result` object given as input does not contain the source distribution.
        """
        if result.forward_speed != 0.0:
            raise NotImplementedError("For free surface elevation with forward speed, please use the `compute_free_surface_elevation` method.")

        fs_elevation = 1j*result.omega/result.g * self.get_potential_on_mesh(result, free_surface.mesh)
        if keep_details:
            result.fs_elevation[free_surface] = fs_elevation
        return fs_elevation



#!/usr/bin/env python
# coding: utf-8
"""Tests for the computation of the Green function using Fortran routines."""

import pytest
from math import log10
import numpy as np
import capytaine as cpt
#----------------------------------------------------------------------------------#
gf = cpt.Delhommeau()

legacy_method = gf.fortran_core.delhommeau_integrals.legacy_method
scaled_nemoh3_method = gf.fortran_core.delhommeau_integrals.scaled_nemoh3_method
'''

for nr in [1000, 1500, 2000]:
    r_range = gf.fortran_core.delhommeau_integrals.default_r_spacing(nr, 100.0, 1)
    print(r_range[(r_range > 0.1) & (r_range < 1.0)])  # Print points between 0.1 and 1

'''
zmin = [-100,-251,-400]
@pytest.mark.parametrize("zmin", zmin)
def test_green_functions_z_spacing_zmin(zmin):
    """Test default_z_spacing"""

    nz = 124
    zRange = gf.fortran_core.delhommeau_integrals.default_z_spacing(nz, zmin, scaled_nemoh3_method)

    dz = []
    for i in range(1,len(zRange)):
        dz.append(log10(abs(zRange[i]))-log10(abs(zRange[i-1])))
    dzComp = dz[-1]*np.ones_like(dz)

    assert abs(zmin-zRange[-1]) < 1e-5
    assert all([abs(a - b) < 1e-5 for a, b in zip(dz, dzComp)])


nz = [64,124,248]
@pytest.mark.parametrize("nz", nz)
def test_green_functions_z_spacing_nz(nz):
    """Test default_z_spacing"""
    zmin = -251
    zRange = gf.fortran_core.delhommeau_integrals.default_z_spacing(nz,zmin, scaled_nemoh3_method)

    dz = []
    for i in range(1,len(zRange)):
        dz.append(log10(abs(zRange[i]))-log10(abs(zRange[i-1])))
    dzComp = dz[-1]*np.ones_like(dz)

    assert abs(zmin-zRange[-1]) < 1e-5
    assert all([abs(a - b) < 1e-5 for a, b in zip(dz, dzComp)])


nr = [100,328,480]
@pytest.mark.parametrize("nr", nr)
def test_green_functions_r_spacing_legacy_nr(nr):
    """Test default_r_spacing legacy model"""

    rmax = 4/3+abs(nr-32)/3
    rRange = gf.fortran_core.delhommeau_integrals.default_r_spacing(nr, rmax, legacy_method)

    r_cal = []
    for i in range(1,nr+1):
        r_cal.append(min(10**((i-1)/5-6),4/3+abs(i-32)/3))

    assert all([abs(a - b) < 1e-4 for a, b in zip(rRange, r_cal)])


rmax = [50,100,200]
@pytest.mark.parametrize("rmax", rmax)
def test_green_functions_r_spacing_Nemoh_rmax(rmax):
    """Test default_r_spacing Nemoh model"""

    nr = 676
    c = 81
    r_logSpace = -np.log(10.0**(-10))/(c - 1.0)
    rRange = gf.fortran_core.delhommeau_integrals.default_r_spacing(nr, rmax, scaled_nemoh3_method)

    r_cal = []
    for i in range(1,nr+1):
        if (i < c):
            r_cal.append(10**(-10)*np.exp(i*r_logSpace))
        else:
            r_cal.append((rmax-1)/(nr-c)*(i-c)+1)

    assert all([abs(a - b) < 1e-4 for a, b in zip(rRange, r_cal)])


nr = [676,1352]
@pytest.mark.parametrize("nr", nr)
def test_green_functions_r_spacing_Nemoh_nr(nr):
    """Test default_r_spacing Nemoh model"""

    nr0 = 676
    c = round(nr/nr0*81)
    rmax = 100
    r_logSpace = -np.log(10.0**(-10))/(c - 1.0)
    rRange = gf.fortran_core.delhommeau_integrals.default_r_spacing(nr, rmax, scaled_nemoh3_method)

    r_cal = []
    for i in range(1,nr+1):
        if (i < c):
            r_cal.append(10**(-10)*np.exp(i*r_logSpace))
        else:
            r_cal.append((rmax-1)/(nr-c)*(i-c)+1)

    assert all([abs(a - b) < 1e-4 for a, b in zip(rRange, r_cal)])




#!/usr/bin/env python

import numpy as np
import capytaine as cpt
from capytaine.bem.airy_waves import airy_waves_free_surface_elevation
import matplotlib.pyplot as plt

mesh = cpt.mesh_sphere(resolution=(10, 10)).immersed_part()
body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs())

pb = cpt.DiffractionProblem(body=body, wave_direction=np.pi/2, wavelength=3.0)
solver = cpt.BEMSolver()
res = solver.solve(pb)

grid = np.meshgrid(np.linspace(-10.0, 10.0, 100), np.linspace(-10.0, 10.0, 100))
fse = solver.compute_free_surface_elevation(grid, res)
incoming_fse = airy_waves_free_surface_elevation(grid, res)

plt.pcolormesh(grid[0], grid[1], np.real(fse + incoming_fse))
plt.xlabel("x")
plt.ylabel("y")
plt.show()



import numpy as np
from functools import wraps

# @total_ordering
class SymbolicMultiplication:
    def __init__(self, symbol, value=1.0):
        self.symbol = symbol
        self.value = value

    def __format__(self, format_spec):
        return f"{self.symbol}×{self.value.__format__(format_spec)}"

    __array_priority__ = 1.0

    def __array_function__(self, func, types, *args, **kwargs):
        if func in {np.real, np.imag, np.sum}:
            return SymbolicMultiplication(self.symbol, func(self.value))
        else:
            return NotImplemented

    def __str__(self):
        return f"{self.symbol}×{self.value}"

    def __repr__(self):
        return f"SymbolicMultiplication(\"{self.symbol}\", {repr(self.value)})"

    def __add__(self, x):
        return self._concretize() + x

    def __radd__(self, x):
        return x + self._concretize()

    def __mul__(self, x):
        return SymbolicMultiplication(self.symbol, self.value * x)

    def __rmul__(self, x):
        return SymbolicMultiplication(self.symbol, x * self.value)

    def __pow__(self, n):
        if n == 2:
            return self * self
        else:
            raise NotImplementedError

    def __truediv__(self, x):
        if hasattr(x, 'symbol') and self.symbol == x.symbol:
            return self.value / x.value
        else:
            return SymbolicMultiplication(self.symbol, self.value / x)

    def __rtruediv__(self, x):
        if hasattr(x, 'symbol') and self.symbol == x.symbol:
            return x.value / self.value
        elif self.symbol == "0":
            return SymbolicMultiplication("∞", x/self.value)
        elif self.symbol == "∞":
            return SymbolicMultiplication("0", x/self.value)
        else:
            raise NotImplementedError

    def __matmul__(self, x):
        return SymbolicMultiplication(self.symbol, self.value @ x)

    def __rmatmul__(self, x):
        return SymbolicMultiplication(self.symbol, x @ self.value)

    def __eq__(self, x):
        return float(self) == x

    def __lt__(self, x):
        return float(self) < x

    def __hash__(self):
        return hash((self.symbol, self.value))

    def _concretize(self):
        if isinstance(self.value, np.ndarray):
            if self.symbol == "0":
                return np.zeros_like(self.value)
            elif self.symbol == "∞":
                return np.full_like(self.value, np.inf)
        else:
            return float(self)

    def __float__(self):
        if self.symbol == "0":
            return 0.0
        elif self.symbol == "∞":
            return np.inf
        else:
            raise NotImplementedError

    def reshape(self, *args):
        return SymbolicMultiplication(self.symbol, self.value.reshape(*args))


def supporting_symbolic_multiplication(f):
    @wraps(f)
    def wrapped_f(a, x):
        if hasattr(x, 'symbol'):
            return SymbolicMultiplication(x.symbol, f(a, x.value))
        else:
            return f(a, x)
    return wrapped_f



import numpy as np
from numpy import pi
import xarray as xr
import matplotlib.pyplot as plt

import capytaine as cpt

cpt.set_logging('INFO')

# SET UP THE MESH
buoy = cpt.FloatingBody(
        mesh=cpt.mesh_vertical_cylinder(
            radius=1.0, length=3.0, center=(0, 0, -1.0),
            resolution=(15, 15, 3)
            ))
buoy.add_translation_dof(name="Surge")

# SOLVE THE BEM PROBLEMS AND COMPUTE THE KOCHIN FUNCTIONS
theta_range = np.linspace(0, 2*pi, 40)
omega_range = np.linspace(1.0, 5.0, 3)

test_matrix = xr.Dataset(coords={
    'omega': omega_range, 'theta': theta_range, 'radiating_dof': ["Surge"],
})
solver = cpt.BEMSolver()
dataset = solver.fill_dataset(test_matrix, buoy.immersed_part(), wavenumber=True)

# PLOT THE KOCHIN FUNCTION
plt.figure()
dataset = dataset.swap_dims({'omega': 'wavenumber'})  # Use wavenumber to index data
for k in dataset.coords['wavenumber']:
    (dataset['kochin']
     .sel(radiating_dof="Surge", wavenumber=k)
     .real
     .plot(x='theta', label=f"k={float(k):.2f} m¯¹", ax=plt.gca()))
plt.legend()
plt.title("Real part of the Kochin function as a function of theta")
plt.tight_layout()
plt.show()



import numpy as np

import capytaine as cpt
import meshmagick.mesh
import meshmagick.hydrostatics

radius = 10
cog = np.array((0, 0, 0))
mesh = cpt.mesh_sphere(radius=radius, center=cog, resolution=(100, 100))
body = cpt.FloatingBody(mesh=mesh, center_of_mass=cog)

body.add_all_rigid_body_dofs()
body = body.immersed_part()

density = 1000
gravity = 9.81

capy_hsdb = body.compute_hydrostatics(rho=density, g=gravity)

stiff_compare_dofs = ["Heave", "Roll", "Pitch"]
capy_hsdb["stiffness_matrix"] = capy_hsdb["hydrostatic_stiffness"].sel(
    influenced_dof=stiff_compare_dofs, radiating_dof=stiff_compare_dofs
    ).values

mass_compare_dofs = ["Roll", "Pitch", "Yaw"]
capy_hsdb["inertia_matrix"] = capy_hsdb["inertia_matrix"].sel(
    influenced_dof=mass_compare_dofs, radiating_dof=mass_compare_dofs
    ).values


mm_mesh = meshmagick.mesh.Mesh(body.mesh.vertices, body.mesh.faces, name=body.mesh.name)

mm_hsdb = meshmagick.hydrostatics.compute_hydrostatics(mm_mesh, cog=cog, rho_water=density, grav=gravity)

mm_hsdb["inertia_matrix"] = mm_mesh.eval_plain_mesh_inertias(rho_medium=density).inertia_matrix
mm_hsdb["center_of_buoyancy"] = mm_hsdb["buoyancy_center"]


analytical = {}
analytical["waterplane_area"] = np.pi*radius**2
analytical["wet_surface_area"] = 2*np.pi*radius**2
analytical["disp_volume"] = (2/3)*np.pi*radius**3
analytical["center_of_buoyancy"] = np.array([0,0,-3*radius/8])
inertia_xx = np.pi*radius**4/4
inertia_yy = np.pi*radius**4/4
inertia_zz = np.pi*radius**4/2
analytical["transversal_metacentric_radius"] = inertia_xx / analytical["disp_volume"]
analytical["longitudinal_metacentric_radius"] = inertia_yy / analytical["disp_volume"]
analytical["transversal_metacentric_height"] = analytical["transversal_metacentric_radius"] + analytical["center_of_buoyancy"][2] - cog[2]
analytical["longitudinal_metacentric_height"] = analytical["longitudinal_metacentric_radius"] + analytical["center_of_buoyancy"][2] - cog[2]

analytical["stiffness_matrix"] = density * gravity * np.array([
    [analytical["waterplane_area"], 0, 0],
    [0, analytical["disp_volume"] * analytical["transversal_metacentric_height"], 0],
    [0, 0, analytical["disp_volume"] * analytical["transversal_metacentric_height"]],
    ])

vars_to_be_displayed = capy_hsdb.keys() & (mm_hsdb.keys() | analytical.keys())
for var in vars_to_be_displayed:
    print(f"{var}:")
    print(f"    Capytaine  :  {capy_hsdb[var]}")
    if var in mm_hsdb:
        print(f"    Meshmagick :  {mm_hsdb[var]}")
    if var in analytical:
        print(f"    Analytical :  {analytical[var]}")



import pytest
import numpy as np
from numpy.random import default_rng
import capytaine as cpt

from capytaine.matrices.block_toeplitz import BlockSymmetricToeplitzMatrix, BlockCirculantMatrix, BlockToeplitzMatrix
from capytaine.matrices.builders import random_block_matrix
from capytaine.matrices.linear_solvers import solve_directly, LUSolverWithCache, solve_gmres

RNG = default_rng(seed=0)

#######################################################################
#                            Full problems                            #
#######################################################################

@pytest.fixture
def solved_full_problem():
    A = RNG.random((5, 5))
    x = RNG.random(A.shape[0])
    return (A, x, A @ x)

def test_solve_directly_full_problem(solved_full_problem):
    A, x_ref, b = solved_full_problem
    x = solve_directly(A, b)
    assert np.allclose(x, x_ref, rtol=1e-10)

def test_solve_with_lu_full_problem(solved_full_problem):
    A, x_ref, b = solved_full_problem
    linear_solver = LUSolverWithCache()
    x = linear_solver.solve(A, b)
    assert np.allclose(x, x_ref, rtol=1e-10)
    x = linear_solver.solve(A, b)  # Should use cache
    assert np.allclose(x, x_ref, rtol=1e-10)

def test_gmres_full_problem(solved_full_problem):
    A, x_ref, b = solved_full_problem
    x = solve_gmres(A, b)
    assert np.allclose(x, x_ref, rtol=1e-10)

#######################################################################
#           2x2 block symmetric matrices (reflection mesh)            #
#######################################################################

@pytest.fixture
def solved_2x2_block_symmetric_toeplitz_problem():
    A = BlockSymmetricToeplitzMatrix([
        [RNG.random((4, 4)) for _ in range(2)]
    ])
    x = RNG.random(A.shape[0])
    return (A, x, A @ x)

def test_solve_directly_2x2_block_symmetric_toeplitz_problem(solved_2x2_block_symmetric_toeplitz_problem):
    A, x_ref, b = solved_2x2_block_symmetric_toeplitz_problem
    x = solve_directly(A, b)
    assert np.allclose(x, x_ref, rtol=1e-10)

def test_solve_with_lu_2x2_block_symmetric_toeplitz_problem(solved_2x2_block_symmetric_toeplitz_problem):
    A, x_ref, b = solved_2x2_block_symmetric_toeplitz_problem
    linear_solver = LUSolverWithCache()
    x = linear_solver.solve(A, b)
    assert np.allclose(x, x_ref, rtol=1e-10)
    x = linear_solver.solve(A, b)  # Should use cache
    assert np.allclose(x, x_ref, rtol=1e-10)

def test_gmres_2x2_block_symmetric_toeplitz_problem(solved_2x2_block_symmetric_toeplitz_problem):
    A, x_ref, b = solved_2x2_block_symmetric_toeplitz_problem
    x = solve_gmres(A, b)
    assert np.allclose(x, x_ref, rtol=1e-10)

#######################################################################
#       Nested 2x2 block symmetric matrices (two reflections)         #
#######################################################################

@pytest.fixture
def solved_nested_2x2_block_symmetric_toeplitz_problem():
    A = BlockSymmetricToeplitzMatrix([
        [BlockSymmetricToeplitzMatrix([
            [RNG.random((4, 4)) for _ in range(2)]
            ]) for _ in range(2)]
        ])
    x = RNG.random(A.shape[0])
    return (A, x, A @ x)

def test_solve_directly_nested_2x2_block_symmetric_toeplitz_problem(solved_nested_2x2_block_symmetric_toeplitz_problem):
    A, x_ref, b = solved_nested_2x2_block_symmetric_toeplitz_problem
    x = solve_directly(A, b)
    assert np.allclose(x, x_ref, rtol=1e-10)

def test_solve_with_lu_nested_2x2_block_symmetric_toeplitz_problem(solved_nested_2x2_block_symmetric_toeplitz_problem):
    A, x_ref, b = solved_nested_2x2_block_symmetric_toeplitz_problem
    linear_solver = LUSolverWithCache()
    x = linear_solver.solve(A, b)
    assert np.allclose(x, x_ref, rtol=1e-10)
    x = linear_solver.solve(A, b)  # Should use cache
    assert np.allclose(x, x_ref, rtol=1e-10)

def test_gmres_2x2_nested_block_symmetric_toeplitz_problem(solved_nested_2x2_block_symmetric_toeplitz_problem):
    A, x_ref, b = solved_nested_2x2_block_symmetric_toeplitz_problem
    x = solve_gmres(A, b)
    assert np.allclose(x, x_ref, rtol=1e-10)

#######################################################################
#                           Block circulant                           #
#######################################################################

@pytest.fixture
def solved_block_circulant_problem():
    A = BlockCirculantMatrix([
        [RNG.random((3, 3)) for _ in range(6)]
    ])
    x = RNG.random(A.shape[0])
    return (A, x, A @ x)

def test_solve_directly_block_circulant_problem(solved_block_circulant_problem):
    A, x_ref, b = solved_block_circulant_problem
    x = solve_directly(A, b)
    assert np.allclose(x, x_ref, rtol=1e-10)

def test_solve_with_lu_block_circulant_problem(solved_block_circulant_problem):
    A, x_ref, b = solved_block_circulant_problem
    linear_solver = LUSolverWithCache()
    with pytest.raises(NotImplementedError):
        linear_solver.solve(A, b)

def test_gmres_block_circulant_problem(solved_block_circulant_problem):
    A, x_ref, b = solved_block_circulant_problem
    x = solve_gmres(A, b)
    assert np.allclose(x, x_ref, rtol=1e-10)


#######################################################################
#                           Block Toeplitz                            #
#######################################################################
@pytest.fixture
def solved_block_toeplitz_problem():
    A = BlockToeplitzMatrix([
        [RNG.random((3, 3)) for _ in range(7)]
    ])
    x = RNG.random(A.shape[0])
    return (A, x, A @ x)

def test_solve_directly_block_toeplitz_problem(solved_block_toeplitz_problem):
    A, x_ref, b = solved_block_toeplitz_problem
    x = solve_directly(A, b)
    assert np.allclose(x, x_ref, rtol=1e-10)

def test_gmres_block_toeplitz_problem(solved_block_toeplitz_problem):
    A, x_ref, b = solved_block_toeplitz_problem
    x = solve_gmres(A, b)
    assert np.allclose(x, x_ref, rtol=1e-10)


#######################################################################
#                            Nested blocks                            #
#######################################################################
@pytest.fixture
def solved_nested_block_problem():
    A = BlockCirculantMatrix([
        [random_block_matrix([1, 1], [1, 1], rng=RNG) for _ in range(6)]
    ])
    x = RNG.random(A.shape[0])
    return (A, x, A @ x)

def test_solve_directly_nested_block_problem(solved_nested_block_problem):
    A, x_ref, b = solved_nested_block_problem
    x = solve_directly(A, b)
    assert np.allclose(x, x_ref, rtol=1e-10)

def test_gmres_nested_block_problem(solved_nested_block_problem):
    A, x_ref, b = solved_nested_block_problem
    x = solve_gmres(A, b)
    assert np.allclose(x, x_ref, rtol=1e-10)



"""
Example file based on Malenica 1995:
Š. Malenica, P.J. Clark, B. Molin,
Wave and current forces on a vertical cylinder free to surge and sway,
Applied Ocean Research, Volume 17, Issue 2, 1995, Pages 79-90,
https://doi.org/10.1016/0141-1187(95)00002-I.

This script is based on the work of the Maritime Technology Division (MTD) of Ghent University, Belgium.
Please refer to following papers:

I. Herdayanditya, L. Donatini, G. Verao Fernandez, A. B. K. Pribadi, and P. Rauwoens,
“Waves-current effect investigation on monopile excitation force employing approximate forward speed approach,”
in 6th MASHCON : international conference on ship manoeuvring in shallow and confined water
with special focus on port manoeuvres, Glasgow, UK, 2022, pp. 73–80.
http://hdl.handle.net/1854/LU-8756871

L. Donatini, I. Herdayanditya, G. Verao Fernandez, A. B. K. Pribadi, E. Lataire, and G. Delefortrie,
“Implementation of forward speed effects on an open source seakeeping solver,”
in 6th MASHCON : international conference on ship manoeuvring in shallow and confined water
with special focus on port manoeuvres, Glasgow, UK, 2022, pp. 20–33.
http://hdl.handle.net/1854/LU-8756868

"""
import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
import capytaine as cpt

g = 9.81
rho = 1025.0

water_depth = 1.0
radius = 1.0
center = (0, 0, -0.5)
resolution = (2, 20, 20)

mesh = cpt.mesh_vertical_cylinder(length=1.001*water_depth, radius=radius, center=center, resolution=resolution)
cylinder = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs(rotation_center=center), center_of_mass=center)
cylinder = cylinder.immersed_part(water_depth=water_depth)

wavenumber_range = np.linspace(0.2,2,25)
froude_number_range = np.array([0, 0.05, -0.05])
forward_speed_range = froude_number_range * np.sqrt(g*radius)

test_matrix = xr.Dataset(coords={
    'wavenumber': wavenumber_range,
    'forward_speed': forward_speed_range,
    'wave_direction': [0.0],
    'radiating_dof': ["Surge"],
    'water_depth': [water_depth],
    'rho': [rho],
    })
solver = cpt.BEMSolver()
dataset = solver.fill_dataset(test_matrix, cylinder)
dataset["excitation_force"] = dataset["Froude_Krylov_force"] + dataset["diffraction_force"]

dataset.coords["encounter_omega"] = dataset["omega"] - dataset["wavenumber"] * dataset["forward_speed"] * np.cos(dataset["wave_direction"])
dataset.coords["Froude_number"] = dataset.coords["forward_speed"]/np.sqrt(g*radius)

plt.figure()
dataset["norm_excitation"] = np.abs(dataset["excitation_force"]/(g*rho*radius**2))
dataset["norm_excitation"].attrs['long_name'] = "Normalized excitation force F/ρgr²"
dataset["norm_excitation"].sel(wave_direction=0.0, influenced_dof="Surge").plot(x="wavenumber", hue="Froude_number")

plt.figure()
dataset["norm_added_mass"] = dataset["added_mass"]/(rho*radius**3)
dataset["norm_added_mass"].attrs['long_name'] = "Normalized added mass A/ρr³"
dataset["norm_added_mass"].sel(influenced_dof="Surge", radiating_dof="Surge").plot(x="wavenumber", hue="Froude_number")

plt.figure()
dataset["norm_rad_damping"] = dataset["radiation_damping"]/(rho*dataset.encounter_omega*radius**3)
dataset["norm_rad_damping"].attrs['long_name'] = "Normalized radiation damping B/ρωr³"
dataset["norm_rad_damping"].sel(influenced_dof="Surge", radiating_dof="Surge").plot(x="wavenumber", hue="Froude_number")

print(dataset)
plt.show()



#!/usr/bin/env python
# coding: utf-8
"""Tests for the mesh submodule: definition and transformation of collections of meshes and symmetric meshes."""

import pytest
import numpy as np
import capytaine as cpt

from capytaine import Disk, AxialSymmetricMesh, HorizontalCylinder, TranslationalSymmetricMesh, CollectionOfMeshes
from capytaine.meshes.meshes import Mesh
from capytaine.bodies.predefined import Sphere
from capytaine.meshes.geometry import xOz_Plane, yOz_Plane


def test_collection_of_meshes():
    # Create some dummy meshes
    vertices = np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]], dtype=float)
    dummy_meshes = [Mesh(vertices, [[0, 1, 2, 3]])]
    for i in range(3):
        dummy_meshes.append(dummy_meshes[0].copy())
        dummy_meshes[i+1].translate_z(i+1)

    # A first collection from a list
    coll = CollectionOfMeshes(dummy_meshes[:2])
    assert coll.nb_submeshes == 2
    assert coll.nb_vertices == 8
    assert coll.nb_faces == 2
    assert coll[1].nb_faces == 1
    assert np.all(coll.faces_areas == np.asarray([1.0, 1.0]))
    assert np.all(coll.faces == np.asarray([[0, 1, 2, 3], [4, 5, 6, 7]]))
    assert np.all(coll.indices_of_mesh(1) == slice(1, 2))

    # A copy of the collection
    copy_coll = coll.copy()
    copy_coll.translate_x(1.0)  # Move
    assert copy_coll.nb_faces == 2
    assert np.all(copy_coll.vertices[:, 0] >= 1.0)  # Has moved

    # Another collection from an iterable
    other_coll = CollectionOfMeshes(iter(dummy_meshes))
    assert other_coll.nb_faces == 4
    assert np.all(other_coll.vertices[:, 0] <= 1.0)  # Did not move

    # A collection of collections
    big_coll = CollectionOfMeshes((copy_coll, other_coll))
    assert big_coll.nb_faces == 6
    assert big_coll.nb_submeshes == 2

    # Move one object in one of the sub-meshes
    copy_coll[1].translate_x(1.0)
    assert big_coll.vertices[:, 0].max() == 3.0

    # Merging the big collection
    merged = big_coll.merged()
    assert isinstance(merged, Mesh)


def test_collection():
    sphere = Sphere(name="foo", center=(0, 0, -2)).mesh
    other_sphere = Sphere(name="bar", center=(0, 0, 2)).mesh

    coll = CollectionOfMeshes([sphere, other_sphere], name="baz")
    coll2 = CollectionOfMeshes([sphere, other_sphere])

    assert coll == coll2
    assert hash(coll) == hash(coll2)

    assert coll[0] == coll2[0]

    coll.tree_view()

    clipped_coll = coll.keep_immersed_part(inplace=False)
    assert len(clipped_coll) == 1
    # assert clipped_coll.merged() == sphere.merged()

    assert np.allclose(sphere.center_of_mass_of_nodes, sphere.merged().center_of_mass_of_nodes)
    assert np.allclose(sphere.diameter_of_nodes, sphere.merged().diameter_of_nodes)


def test_join_axisymmetric_disks():
    """Given two axisymmetric meshes with the same axis, build a new axisymmetric mesh combining the two."""
    disk1 = Disk(radius=1.0, center=(-1, 0, 0), resolution=(6, 6), normal=(1, 0, 0), axial_symmetry=True).mesh
    disk2 = Disk(radius=2.0, center=(1, 0, 0), resolution=(8, 6), normal=(1, 0, 0), axial_symmetry=True).mesh
    joined = disk1.join_meshes(disk2, name="two_disks")
    assert isinstance(joined, AxialSymmetricMesh)
    joined.tree_view()

    disk3 = Disk(radius=1.0, center=(0, 0, 0), resolution=(6, 4), axial_symmetry=True).mesh
    with pytest.raises(AssertionError):
        disk1.join_meshes(disk3)


def test_join_translational_cylinders():
    """Given two meshes with the same translation symmetry, join them into a single mesh with the same symmetry."""
    params = dict(length=10.0, reflection_symmetry=False, translation_symmetry=True, nr=0, ntheta=10, nx=10)
    mesh1 = HorizontalCylinder(radius=1.0, center=(0, 5, -5), **params).mesh
    mesh2 = HorizontalCylinder(radius=2.0, center=(0, -5, -5), **params).mesh
    joined = mesh1.join_meshes(mesh2)
    assert isinstance(joined, TranslationalSymmetricMesh)


def test_mesh_splitting():
    mesh = Sphere().mesh.merged()

    splitted_mesh = mesh.sliced_by_plane(xOz_Plane.translated_y(0.5))
    assert isinstance(splitted_mesh, CollectionOfMeshes)
    assert splitted_mesh.merged() == mesh

    twice_splitted_mesh = splitted_mesh.sliced_by_plane(yOz_Plane)
    assert isinstance(twice_splitted_mesh[0], CollectionOfMeshes)
    assert isinstance(twice_splitted_mesh[1], CollectionOfMeshes)
    assert twice_splitted_mesh.merged() == mesh


def test_extract_one_face():
    sphere = Sphere().mesh
    assert sphere.submesh_containing_face(0) == (0, 0)
    assert sphere.submesh_containing_face(1) == (0, 1)
    assert sphere.submesh_containing_face(sphere[0].nb_faces-1) == (0, sphere[0].nb_faces-1)
    assert sphere.submesh_containing_face(sphere[0].nb_faces) == (1, 0)
    assert sphere.submesh_containing_face(sphere[0].nb_faces+1) == (1, 1)

    for i in [5, sphere[0].nb_faces+5]:
        assert np.allclose(sphere.extract_one_face(i).faces_centers[0], sphere.faces_centers[i])


def test_immersed_part():
    mesh = cpt.mesh_horizontal_cylinder(reflection_symmetry=True)
    assert mesh.immersed_part().merged() == mesh.merged().immersed_part()


def test_path_to_leaf():
    sphere_1 = cpt.mesh_sphere()
    sphere_2 = cpt.mesh_sphere()
    sphere_3 = cpt.mesh_sphere()
    sphere_4 = cpt.mesh_sphere()
    coll = cpt.CollectionOfMeshes([
        cpt.CollectionOfMeshes([sphere_1, sphere_2]),
        sphere_3, sphere_4])
    assert sphere_1.path_to_leaf() == [[]]
    assert coll.path_to_leaf() == [[0, 0], [0, 1], [1], [2]]



import pytest
from pytest import approx
import numpy as np
import pandas as pd
import capytaine as cpt
import xarray as xr

# PROBLEM DEFINITION

@pytest.fixture
def body():
    mesh = cpt.mesh_vertical_cylinder(radius=1.0, length=1.01, center=(0.0, 0.0, -0.5), resolution=(2, 20, 10)).immersed_part(water_depth=1.0)
    body = cpt.FloatingBody(mesh=mesh, name="body")
    body.add_translation_dof(name="Surge")
    return body

@pytest.fixture
def solver():
    return cpt.BEMSolver()

def test_encouter_frequency_along_waves(body):
    pb = cpt.DiffractionProblem(body=body, omega=2.0, forward_speed=1.0, wave_direction=0.0)
    assert pb.forward_speed == 1.0
    assert pb.encounter_omega < pb.omega  # Object is moving in the same direction as the waves
    assert pb.encounter_wave_direction == pb.wave_direction

def test_encouter_frequency_against_waves(body):
    pb = cpt.DiffractionProblem(body=body, omega=2.0, forward_speed=-1.0, wave_direction=0.0)
    assert pb.forward_speed == -1.0
    assert pb.encounter_omega > pb.omega  # Object is moving against the wave
    assert pb.encounter_wave_direction == pb.wave_direction

def test_encouter_frequency_faster_than_waves(body):
    pb = cpt.DiffractionProblem(body=body, omega=2.0, forward_speed=10.0, wave_direction=0.0)
    assert pb.forward_speed == 10.0
    assert pb.encounter_omega >= 0.0
    assert pb.encounter_wave_direction == approx(pb.wave_direction + np.pi)

def test_encouter_frequency_orthogonal_waves(body):
    pb = cpt.DiffractionProblem(body=body, omega=2.0, forward_speed=1.0, wave_direction=np.pi/2)
    assert pb.forward_speed == 1.0
    assert pb.encounter_omega == approx(pb.omega)

def test_encounter_frequency_radiation_problem(body):
    pb = cpt.RadiationProblem(body=body, omega=2.0, forward_speed=1.0, wave_direction=0.0, radiating_dof="Surge")
    assert pb.forward_speed == 1.0
    assert pb.encounter_omega < pb.omega

def test_multibody(body, solver):
    two_bodies = body + body.translated_x(5.0, name="other_body")
    with pytest.raises(NotImplementedError):
        pb = cpt.RadiationProblem(body=two_bodies, omega=2.0, forward_speed=1.0, radiating_dof="body__Surge")

def test_non_rigid_body(body, solver):
    body = body.copy(name="body")
    body.dofs["Shear"] = np.array([[z, 0, 0] for (x, y, z) in body.mesh.faces_centers])
    with pytest.raises(NotImplementedError):
        pb = cpt.RadiationProblem(body=body, omega=2.0, forward_speed=1.0, radiating_dof="Shear")

# POST-PROCESSING

@pytest.fixture
def result(body, solver):
    pb = cpt.DiffractionProblem(body=body, omega=2.0, forward_speed=1.0, wave_direction=0.0)
    return solver.solve(pb)

def test_pressure_reconstruction(result, solver):
    pressure = solver.compute_pressure(result.body.mesh, result)
    assert result.body.integrate_pressure(pressure) == approx(result.forces)

def test_velocity_reconstruction(result, solver):
    points = np.meshgrid(np.linspace(-10.0, 10.0, 3), np.linspace(-10.0, 10.0, 3), np.linspace(-10.0, -1.0, 2))
    velocity = solver.compute_velocity(points, result)

def test_free_surface_elevation(result, solver):
    points = np.meshgrid(np.linspace(-10.0, 10.0, 3), np.linspace(-10.0, 10.0, 3))
    fse = solver.compute_free_surface_elevation(points, result)

# DATASETS

def test_problem_from_dataset(body, solver):
    from capytaine.io.xarray import problems_from_dataset
    test_matrix = xr.Dataset(coords={
        "omega": [1.0],
        "forward_speed": [0.0, 1.0],
        "wave_direction": [0.0, np.pi/2],
        "radiating_dof": ["Surge"],
        })
    pbs = problems_from_dataset(test_matrix, body)
    assert len(pbs) == 7
    # Four diffraction problems + Three radiation problems, since the wave_direction is only relevant when forward_speed is non 0.0

def test_fill_dataset_with_forward_speed(body, solver):
    test_matrix = xr.Dataset(coords={
        "omega": [1.0],
        "forward_speed": [0.0, 1.0],
        "wave_direction": [0.0, np.pi/2],
        "radiating_dof": ["Surge"],
        })
    ds = solver.fill_dataset(test_matrix, body)
    assert "wave_direction" in ds.added_mass.coords
    assert "encounter_omega" in ds.coords
    assert "encounter_wave_direction" in ds.coords

def test_fill_dataset_without_forward_speed(body, solver):
    test_matrix = xr.Dataset(coords={
        "omega": [1.0],
        "wave_direction": [0.0, np.pi/2],
        "radiating_dof": ["Surge"],
        })
    ds = solver.fill_dataset(test_matrix, body)
    assert "wave_direction" not in ds.added_mass.coords
    assert "encounter_omega" not in ds.coords
    assert "encounter_wave_direction" not in ds.coords

def test_no_explicit_wave_direction(body):
    test_matrix = xr.Dataset(coords={
        "wavelength": np.linspace(0.1, 10.0, 2),
        "forward_speed": [0.0, 10.0],
        "radiating_dof": ["Surge"],
        })
    ds = cpt.BEMSolver().fill_dataset(test_matrix, body)
    assert ds.wave_direction.values == np.array([0.0])


# VALIDATION

rho = 1025
g = 9.81

# Based on Figure 1 from Malenica 1995
MALENICA_EXCITATION_FORCE = pd.DataFrame([
    (0.269, 1.662, 0.05), (0.615, 3.192, 0.05), (1.086, 3.168, 0.05), (1.568, 2.225, 0.05), (1.955, 1.608, 0.05),
    (0.201, 1.281, 0.0), (0.506, 2.928, 0.0), (1.308, 2.733, 0.0), (1.840, 1.890, 0.0),
    (0.353, 2.263, -0.05), (0.568, 3.224, -0.05), (1.171, 2.963, -0.05), (1.977, 1.893, -0.05),
    ], columns=["wavenumber", "normalized_force", "froude_number"])

@pytest.mark.parametrize("ref_data", MALENICA_EXCITATION_FORCE.iterrows())
def test_malenica_excitation_force(body, solver, ref_data):
    from capytaine.bem.airy_waves import froude_krylov_force
    pb = cpt.DiffractionProblem(
        body=body,
        wavenumber=ref_data[1].wavenumber,
        forward_speed=ref_data[1].froude_number*np.sqrt(g),
        wave_direction=0.0,
        water_depth=1.0,
        rho=rho
    )
    res = solver.solve(pb)
    excitation_force = res.forces["Surge"] + froude_krylov_force(pb)["Surge"]
    assert np.abs(excitation_force)/(g*rho) == approx(ref_data[1].normalized_force, rel=5e-2)


MALENICA_ADDED_MASS = pd.DataFrame([
    (0.620, 3.302, 0.05), (0.855, 2.634, 0.05), (1.303, 1.407, 0.05),
    (0.510, 3.433, 0.0), (0.916, 2.217, 0.0), (1.495, 0.916, 0.0),
    (0.602, 3.172, -0.05), (1.069, 1.518, -0.05),
    ], columns=["wavenumber", "normalized_added_mass", "froude_number"])

@pytest.mark.parametrize("ref_data", MALENICA_ADDED_MASS.iterrows())
def test_malenica_added_mass(body, solver, ref_data):
    from capytaine.bem.airy_waves import froude_krylov_force
    pb = cpt.RadiationProblem(
        body=body,
        wavenumber=ref_data[1].wavenumber,
        forward_speed=ref_data[1].froude_number*np.sqrt(g),
        wave_direction=0.0,
        water_depth=1.0,
        radiating_dof="Surge",
        rho=rho
    )
    res = solver.solve(pb)
    assert np.abs(res.added_masses["Surge"])/(rho) == approx(ref_data[1].normalized_added_mass, rel=1e-1)

MALENICA_RADIATION_DAMPING = pd.DataFrame([
    (0.456, 0.965, 0.05), (0.727, 1.917, 0.05), (1.129, 2.284, 0.05),
    (0.434, 0.978, 0.0), (0.674, 1.884, 0.0), (1.125, 2.218, 0.0),
    (0.408, 0.971, -0.05), (0.642, 1.906, -0.05), (1.042, 2.234, -0.05),
    ], columns=["wavenumber", "normalized_radiation_damping", "froude_number"])

@pytest.mark.parametrize("ref_data", MALENICA_RADIATION_DAMPING.iterrows())
def test_malenica_radiation_damping(body, solver, ref_data):
    from capytaine.bem.airy_waves import froude_krylov_force
    pb = cpt.RadiationProblem(
        body=body,
        wavenumber=ref_data[1].wavenumber,
        forward_speed=ref_data[1].froude_number*np.sqrt(g),
        wave_direction=0.0,
        water_depth=1.0,
        radiating_dof="Surge",
        rho=rho
    )
    res = solver.solve(pb)
    assert np.abs(res.radiation_dampings["Surge"])/(rho*pb.encounter_omega) == approx(ref_data[1].normalized_radiation_damping, rel=1e-1)

    
def test_near_zero_encounter_frequency_radiation(body, solver):
    pb = cpt.RadiationProblem(body=body, omega=1.0, forward_speed=9.805, radiating_dof="Surge")
    assert pb.encounter_omega > 0.0
    solver.solve(pb)


def test_near_zero_encounter_frequency_diffraction(body, solver):
    pb = cpt.DiffractionProblem(body=body, omega=1.0, forward_speed=9.805)
    assert pb.encounter_omega > 0.0
    solver.solve(pb)

    
def test_zero_encounter_frequency_radiation(body, solver):
    pb = cpt.RadiationProblem(body=body, omega=1.0, forward_speed=9.81, radiating_dof="Surge")
    assert float(pb.encounter_omega) == 0.0
    res = solver.solve(pb)
    assert not np.isinf(res.forces["Surge"])
    assert not np.isnan(res.forces["Surge"])
    assert np.isinf(res.added_mass["Surge"])

    
def test_zero_encounter_frequency_diffraction(body, solver):
    pb = cpt.DiffractionProblem(body=body, omega=1.0, forward_speed=9.81)
    assert float(pb.encounter_omega) == 0.0
    res = solver.solve(pb)
    assert not np.isinf(res.forces["Surge"])
    assert not np.isnan(res.forces["Surge"])


"""Importing mesh from meshio"""
# Copyright (C) 2017-2022 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>
import logging

import numpy as np

from capytaine.tools.optional_imports import import_optional_dependency
from capytaine.meshes.meshes import Mesh

LOG = logging.getLogger(__name__)

def load_from_meshio(mesh, name=None):
    """Create a Mesh from a meshio mesh object."""
    meshio = import_optional_dependency("meshio")
    if not isinstance(mesh, meshio._mesh.Mesh):
        raise TypeError('mesh must be of type meshio._mesh.Mesh, received {:}'.format(type(mesh)))

    def all_faces_as_quads(cells):
        all_faces = []
        if 'quad' in cells:
            all_faces.append(cells['quad'])
        if 'triangle' in cells:
            num_triangles = len(mesh.cells_dict['triangle'])
            LOG.info("Stored {:} triangle faces as quadrilaterals".format(num_triangles))
            triangles_as_quads = np.empty((cells['triangle'].shape[0], 4), dtype=int)
            triangles_as_quads[:, :3] = cells['triangle'][:, :]
            triangles_as_quads[:, 3] = cells['triangle'][:, 2]  # Repeat one node to make a quad
            all_faces.append(triangles_as_quads)
        return np.concatenate(all_faces)

    if name is None:
        name = f'mesh_from_meshio_{next(Mesh._ids)}'

    return Mesh(vertices=mesh.points, faces=all_faces_as_quads(mesh.cells_dict), name=name)



# Copyright (C) 2017-2024 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>
"""Tools for memoization of functions."""
from collections import OrderedDict
from functools import wraps

import logging

LOG = logging.getLogger(__name__)


def lru_cache_with_strict_maxsize(maxsize=1):
    """Behaves mostly like functools.lru_cache(), but the oldest data in the cache is
    deleted *before* computing a new one, in order to *never* have more that
    `maxsize` items in memory.
    This is useful to limit RAM usage when stored objects are big, like the interaction
    matrices of Capytaine."""

    def decorator(f):
        cache = OrderedDict()

        @wraps(f)
        def decorated_f(*args, **kwargs):
            hashable_kwargs = tuple((k, v) for (k, v) in kwargs.items())
            # Might miss a cache hit if the order of kwargs is changed.
            # But at least unlike a previous version, should not return a wrong value.

            if (args, hashable_kwargs) in cache:
                # Get item in cache
                LOG.debug("Get cached version of %s(%s, %s)", f.__name__, args, hashable_kwargs)
                return cache[(args, hashable_kwargs)]

            if len(cache) + 1 > maxsize:
                # Drop oldest item in cache.
                cache.popitem(last=False)

            # Compute and store
            LOG.debug("Computing %s(%s, %s)", f.__name__, args, hashable_kwargs)
            result = f(*args, **kwargs)
            cache[(args, hashable_kwargs)] = result

            return result

        return decorated_f

    return decorator


delete_first_lru_cache = lru_cache_with_strict_maxsize  # For backward compatibility...



"""
Adapted from https://github.com/platformdirs/platformdirs (MIT Licensed)
"""
import os
import sys
from pathlib import Path

from capytaine import __version__


def cache_directory():
    if sys.platform == "win32":  # Windows
        path = os.path.normpath(os.environ.get("LOCALAPPDATA"))
        path = os.path.join(path, "capytaine", "Cache", __version__)
    elif sys.platform == "darwin":  # MacOS
        path = os.path.expanduser("~/Library/Caches")
        path = os.path.join(path, "capytaine", __version__)
    else:
        path = os.environ.get("XDG_CACHE_HOME", "")
        if path.strip() == "":
            path = os.path.expanduser("~/.cache")
        path = os.path.join(path, "capytaine", __version__)
    Path(path).mkdir(parents=True, exist_ok=True)
    return path



import pytest
import numpy as np
import xarray as xr
import capytaine as cpt


@pytest.fixture
def sphere_fb():
    m = 1.866e+03
    mesh = cpt.mesh_sphere(radius=1.0, resolution=(3, 12)).immersed_part()
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs(), mass=m)
    body.inertia_matrix = body.add_dofs_labels_to_matrix(
            [[ m ,  0.0, 0.0, 0.0,        0.0,       0.0       ],
             [ 0.0, m,   0.0, 0.0,        0.0,       0.0       ],
             [ 0.0, 0.0, m,   0.0,        0.0,       0.0       ],
             [ 0.0, 0.0, 0.0, 4.469e+02,  9.676e-31, -2.757e-14],
             [ 0.0, 0.0, 0.0, 9.676e-31,  4.469e+02, 3.645e-15 ],
             [ 0.0, 0.0, 0.0, -2.757e-14, 3.645e-15, 6.816e+02 ]]
            )
    body.hydrostatic_stiffness = body.add_dofs_labels_to_matrix(
            [[    0.   ,     0.   ,     0.   ,     0.   ,     0.   ,     0.   ],
             [    0.   ,     0.   ,     0.   ,     0.   ,     0.   ,     0.   ],
             [    0.   ,     0.   , 29430.   ,     0.   ,     0.   ,     0.   ],
             [    0.   ,     0.   ,     0.   ,   328.573,     0.   ,     0.   ],
             [    0.   ,     0.   ,     0.   ,     0.   ,   328.573,     0.   ],
             [    0.   ,     0.   ,     0.   ,     0.   ,     0.   ,     0.   ]]
            )
    return body

@pytest.fixture
def solver():
    return cpt.BEMSolver()

def test_rao_sphere_all(sphere_fb, solver):
    test_matrix = xr.Dataset(coords={
        'omega': np.linspace(0.5, 10.0, 5),
        'wave_direction': [0],
        'radiating_dof': list(sphere_fb.dofs.keys()),
        })

    data = solver.fill_dataset(test_matrix, sphere_fb,
                               hydrostatics=True, mesh=True)


    RAO = cpt.post_pro.rao(data)

    assert RAO.radiating_dof.size == 6
    assert data.inertia_matrix.shape == (6,6)
    assert np.all(data.inertia_matrix.values == sphere_fb.inertia_matrix.values)
    assert data.hydrostatic_stiffness.shape == (6,6)
    assert np.all(data.hydrostatic_stiffness.values == sphere_fb.hydrostatic_stiffness.values)
    # # assert RAO == ? # TODO could test against known results


def test_rao_from_wavelengths(sphere_fb, solver):
    # From https://github.com/capytaine/capytaine/issues/316
    test_matrix = xr.Dataset(coords={
        'wavelength': np.linspace(0.5, 10.0, 5),
        'wave_direction': [0],
        'radiating_dof': list(sphere_fb.dofs.keys()),
        })

    data = solver.fill_dataset(test_matrix, sphere_fb, hydrostatics=True)
    Zi = cpt.post_pro.impedance(data)
    RAO = cpt.post_pro.rao(data)


def test_rao_several_water_depth(sphere_fb, solver):
    # From https://github.com/capytaine/capytaine/issues/405
    test_matrix = xr.Dataset(coords={
        'omega': np.linspace(0.5, 10.0, 3),
        'wave_direction': [0],
        'water_depth': [np.inf, 10.0],
        'radiating_dof': list(sphere_fb.dofs.keys()),
        })

    data = solver.fill_dataset(test_matrix, sphere_fb, hydrostatics=True)
    rao = cpt.post_pro.rao(data)
    assert "water_depth" in rao.dims


@pytest.fixture
def sphere_heave_data(sphere_fb, solver):
    sphere_fb.keep_only_dofs(['Heave'])

    test_matrix = xr.Dataset(coords={
          'omega': np.linspace(0.5, 10.0, 5),
          'wave_direction': [0],
          'radiating_dof': list(sphere_fb.dofs.keys()),
          })

    data = solver.fill_dataset(test_matrix, [sphere_fb], hydrostatics=True)

    return data


def test_impedance_sphere_heave(sphere_heave_data):
    Zi = cpt.post_pro.impedance(sphere_heave_data)


def test_rao_sphere_heave_indirect(sphere_heave_data):
    RAO = cpt.post_pro.rao(sphere_heave_data)
    assert RAO.radiating_dof.size == 1
    assert sphere_heave_data.inertia_matrix.size == 1
    assert sphere_heave_data.hydrostatic_stiffness.size == 1
    # assert RAO == ? # TODO could test against known results



import os
import logging

import pytest

import numpy as np
import xarray as xr

import capytaine as cpt

from capytaine.bem.problems_and_results import LinearPotentialFlowProblem, \
        LinearPotentialFlowResult, DiffractionResult, RadiationResult
from capytaine.io.xarray import problems_from_dataset, assemble_dataset
from capytaine.io.legacy import import_cal_file


@pytest.fixture
def sphere():
    sphere = cpt.FloatingBody(
            mesh=cpt.mesh_sphere(center=(0, 0, -2), radius=1.0, resolution=(10, 20)),
            name="sphere",
            )
    sphere.add_translation_dof(direction=(0, 0, 1), name="Heave")
    return sphere


@pytest.fixture
def solver():
    solver = cpt.BEMSolver()
    return solver


def test_LinearPotentialFlowProblem(sphere):
    # Without a body
    pb = LinearPotentialFlowProblem(omega=1.0)
    assert pb.omega == 1.0
    assert pb.period == 2*np.pi
    assert pb.wavenumber == 1.0/9.81
    assert pb.wavelength == 9.81*2*np.pi

    assert LinearPotentialFlowProblem(free_surface=np.inf, water_depth=np.inf).water_depth == np.inf
    assert LinearPotentialFlowProblem(free_surface=0.0, water_depth=np.inf).water_depth == np.inf

    pb = LinearPotentialFlowProblem(free_surface=0.0, water_depth=1.0, omega=1.0)
    assert pb.water_depth == 1.0
    assert np.isclose(pb.omega**2, pb.g*pb.wavenumber*np.tanh(pb.wavenumber*pb.water_depth))

    with pytest.raises(NotImplementedError):
        LinearPotentialFlowProblem(free_surface=2.0)

    with pytest.raises(NotImplementedError):
        LinearPotentialFlowProblem(free_surface=np.inf, water_depth=2.0)

    with pytest.raises(ValueError):
        LinearPotentialFlowProblem(free_surface=0.0, water_depth=-1.0)

    with pytest.raises(TypeError):
        LinearPotentialFlowProblem(radiating_dof="Heave")

    with pytest.raises(ValueError):
        LinearPotentialFlowProblem(body=cpt.FloatingBody(mesh=cpt.Mesh([], [])))

    # With a body
    pb = LinearPotentialFlowProblem(body=sphere, omega=1.0)
    pb.boundary_condition = sphere.mesh.faces_normals @ (1, 1, 1)
    assert list(pb.influenced_dofs.keys()) == ['Heave']

    pb2 = LinearPotentialFlowProblem(body=sphere, omega=2.0)
    pb2.boundary_condition = sphere.mesh.faces_normals @ (1, 1, 1)
    assert pb < pb2

    # Test transformation to result class
    res = pb.make_results_container()
    assert isinstance(res, LinearPotentialFlowResult)
    assert res.problem is pb
    assert res.omega == pb.omega
    assert res.period == pb.period
    assert res.body is pb.body


def test_mesh_above_the_free_surface(caplog):
    with caplog.at_level(logging.WARNING):
        body = cpt.FloatingBody(mesh=cpt.mesh_sphere(), dofs=cpt.rigid_body_dofs())
        pb = cpt.RadiationProblem(body=body, omega=1.0)
    assert '50 panels above the free surface' in caplog.text
    assert np.all(pb.body.mesh.vertices[:, 2].max() <= 0.0)


def test_mesh_below_the_sea_bottom(caplog):
    with caplog.at_level(logging.WARNING):
        body = cpt.FloatingBody(
                mesh=cpt.mesh_sphere(radius=0.5, center=(0, 0, -1.0)),
                dofs=cpt.rigid_body_dofs()
                )
        pb = cpt.RadiationProblem(body=body, omega=1.0, water_depth=1.0)
    assert '50 panels below the sea bottom' in caplog.text
    assert np.all(pb.body.mesh.vertices[:, 2].min() >= -1.0)


def test_backward_compatibility_with_sea_bottom_argument(caplog):
    with caplog.at_level(logging.WARNING):
        pb = cpt.DiffractionProblem(sea_bottom=-10.0)
        assert pb.water_depth == 10.0
    assert 'water_depth' in caplog.text


@pytest.mark.parametrize("water_depth", [10.0, np.inf])
def test_setting_wavelength(water_depth):
    λ = 10*np.random.rand()
    assert np.isclose(cpt.DiffractionProblem(wavelength=λ, water_depth=water_depth).wavelength, λ)


@pytest.mark.parametrize("water_depth", [10.0, np.inf])
def test_setting_wavenumber(water_depth):
    k = 10*np.random.rand()
    assert np.isclose(cpt.DiffractionProblem(wavenumber=k, water_depth=water_depth).wavenumber, k)


@pytest.mark.parametrize("water_depth", [10.0, np.inf])
def test_setting_period(water_depth):
    T = 10*np.random.rand()
    assert np.isclose(cpt.DiffractionProblem(period=T, water_depth=water_depth).period, T)


def test_setting_too_many_frequencies():
    with pytest.raises(ValueError, match="at most one"):
        cpt.DiffractionProblem(omega=1.0, wavelength=1.0)


def test_diffraction_problem(sphere):
    assert cpt.DiffractionProblem(omega=1.0).body is None

    pb = cpt.DiffractionProblem(body=sphere, wave_direction=1.0, wavenumber=1.0)
    assert len(pb.boundary_condition) == sphere.mesh.nb_faces

    with pytest.raises(TypeError):
        cpt.DiffractionProblem(boundary_conditions=[0, 0, 0])

    assert "DiffractionProblem" in str(cpt.DiffractionProblem(wavenumber=1.0, g=10, rho=1025, free_surface=np.inf))

    res = pb.make_results_container()
    assert isinstance(res, DiffractionResult)


def test_wave_direction_radians_warning(sphere, caplog):
    with caplog.at_level(logging.WARNING):
        cpt.DiffractionProblem(body=sphere, omega=1.0, wave_direction=180)
    assert 'in radians and not in degrees' in caplog.text


def test_radiation_problem(sphere, caplog):
    pb = cpt.RadiationProblem(body=sphere, omega=1.0)
    assert len(pb.boundary_condition) == sphere.mesh.nb_faces

    sphere.add_translation_dof(direction=(1, 0, 0), name="Surge")
    pb2 = cpt.RadiationProblem(body=sphere, radiating_dof="Heave")
    assert np.all(pb.boundary_condition == pb2.boundary_condition)

    assert "RadiationProblem" in str(cpt.RadiationProblem(g=10, rho=1025, free_surface=np.inf))

    res = pb.make_results_container()
    assert isinstance(res, RadiationResult)
    assert res.added_mass == res.added_masses == {}
    assert res.radiation_damping == res.radiation_dampings == {}

    res = pb.make_results_container(forces={"Heave": 1.0 + 2.0j})
    assert res.added_mass == res.added_masses == {"Heave": 1.0}
    assert res.radiation_damping == res.radiation_dampings == {"Heave": 2.0}


@pytest.mark.parametrize("cal_file", ["Nemoh.cal", "Nemoh_v3.cal"])
def test_import_cal_file(cal_file):
    """Test the importation of legacy Nemoh.cal files."""
    current_file_path = os.path.dirname(os.path.abspath(__file__))

    # Non symmetrical body
    cal_file_path = os.path.join(current_file_path, "Nemoh_verification_cases", "NonSymmetrical", cal_file)
    problems = import_cal_file(cal_file_path)

    assert len(problems) == 6*41+41
    for problem in problems:
        assert problem.rho == 1000.0
        assert problem.g == 9.81
        assert problem.water_depth == np.inf
        assert isinstance(problem.body, cpt.FloatingBody)
        assert problem.body.nb_dofs == 6
        assert problem.body.mesh.nb_vertices == 299  # Duplicate vertices are removed during import.
        assert problem.body.mesh.nb_faces == 280
        assert problem.omega in np.linspace(0.1, 2.0, 41)
        if isinstance(problem, cpt.DiffractionProblem):
            assert problem.wave_direction == 0.0

    # Symmetrical cylinder
    cal_file_path = os.path.join(current_file_path, "Nemoh_verification_cases", "Cylinder", cal_file)
    problems = import_cal_file(cal_file_path)

    assert len(problems) == 6*2+2
    for problem in problems:
        assert problem.rho == 1000.0
        assert problem.g == 9.81
        assert problem.water_depth == np.inf
        assert isinstance(problem.body.mesh, cpt.ReflectionSymmetricMesh)
        assert isinstance(problem.body.mesh[0], cpt.Mesh)
        assert problem.body.nb_dofs == 6
        # assert problem.body.mesh.nb_vertices == 2*540
        assert problem.body.mesh.nb_faces == 2*300
        assert problem.omega == 0.1 or problem.omega == 2.0
        if isinstance(problem, cpt.DiffractionProblem):
            assert problem.wave_direction == 0.0


def test_results():
    assert isinstance(LinearPotentialFlowProblem().make_results_container(), LinearPotentialFlowResult)

    pb = cpt.DiffractionProblem(g=10, rho=1023, free_surface=np.inf)
    res = DiffractionResult(pb)
    assert res.g == pb.g == 10
    assert "DiffractionResult" in str(res)

    pb = cpt.RadiationProblem(g=10, rho=1023, free_surface=np.inf)
    res = RadiationResult(pb)
    assert res.g == pb.g == 10
    assert "RadiationResult" in str(res)


def test_problems_from_dataset(sphere):
    dset = xr.Dataset(coords={'omega': [0.5, 1.0, 1.5],
                              'radiating_dof': ["Heave"],
                              'body_name': ["sphere"],
                              'wave_direction': [0.0],
                              'water_depth': [np.inf]})

    problems = problems_from_dataset(dset, [sphere])
    assert cpt.RadiationProblem(body=sphere, omega=0.5, radiating_dof="Heave") in problems
    assert len(problems) == 6
    assert len([problem for problem in problems if isinstance(problem, cpt.DiffractionProblem)]) == 3

    dset = xr.Dataset(coords={'omega': [0.5, 1.0, 1.5],
                              'wave_direction': [0.0],
                              'body_name': ["cube"]})
    with pytest.raises(AssertionError):
        problems_from_dataset(dset, [sphere])

    shifted_sphere = sphere.translated_y(5.0, name="shifted_sphere")
    dset = xr.Dataset(coords={'omega': [0.5, 1.0, 1.5],
                              'radiating_dof': ["Heave"],
                              'wave_direction': [0.0]})
    problems = problems_from_dataset(dset, [sphere, shifted_sphere])
    assert cpt.RadiationProblem(body=sphere, omega=0.5, radiating_dof="Heave") in problems
    assert cpt.RadiationProblem(body=shifted_sphere, omega=0.5, radiating_dof="Heave") in problems
    assert len(problems) == 12


def test_problems_from_dataset_with_wavelength(sphere):
    dset = xr.Dataset(coords={'wavelength': [12.0], 'radiating_dof': ["Heave"]})
    problems = problems_from_dataset(dset, sphere)
    for pb in problems:
        assert np.isclose(pb.wavelength, 12.0)


def test_problems_from_dataset_with_too_many_info(sphere):
    dset = xr.Dataset(coords={'wavelength': [12.0], 'period': [3.0], 'radiating_dof': ["Heave"]})
    with pytest.raises(ValueError, match="at most one"):
        problems = problems_from_dataset(dset, sphere)


@pytest.mark.parametrize("method", ['indirect','direct'])
def test_assemble_dataset(sphere, solver, method):
    pb_1 = cpt.DiffractionProblem(body=sphere, wave_direction=1.0, omega=1.0)
    res_1 = solver.solve(pb_1, method=method)
    ds1 = assemble_dataset([res_1])
    assert "Froude_Krylov_force" in ds1

    pb_2 = cpt.RadiationProblem(body=sphere, radiating_dof="Heave", omega=1.0)
    res_2 = solver.solve(pb_2, method=method)
    ds2 = assemble_dataset([res_2])
    assert "added_mass" in ds2

    ds12 = assemble_dataset([res_1, res_2])
    assert "Froude_Krylov_force" in ds12
    assert "added_mass" in ds12


def test_fill_dataset(sphere, solver):
    sphere.add_all_rigid_body_dofs()
    test_matrix = xr.Dataset(coords={'omega': [1.0, 2.0, 3.0], 'wave_direction': [0, np.pi/2], 'radiating_dof': ['Heave']})
    dataset = solver.fill_dataset(test_matrix, [sphere])
    assert dataset['added_mass'].data.shape == (3, 1, 6)
    assert dataset['Froude_Krylov_force'].data.shape == (3, 2, 6)


def test_fill_dataset_with_wavenumbers(sphere, solver):
    k_range = np.linspace(1.0, 3.0, 3)
    test_matrix = xr.Dataset(coords={'wavenumber': k_range, 'wave_direction': [0, np.pi/2], 'radiating_dof': ['Heave']})
    dataset = solver.fill_dataset(test_matrix, [sphere])
    np.testing.assert_allclose(dataset.coords['wavenumber'], k_range)
    assert set(dataset.added_mass.dims) == {'wavenumber', 'radiating_dof', 'influenced_dof'}
    assert set(dataset.wavenumber.dims) == {'wavenumber'}
    assert set(dataset.wavelength.dims) == {'wavenumber'}
    assert set(dataset.omega.dims)      == {'wavenumber'}
    assert set(dataset.period.dims)     == {'wavenumber'}


def test_fill_dataset_with_periods(sphere, solver):
    T_range = np.linspace(1.0, 3.0, 3)
    test_matrix = xr.Dataset(coords={'period': T_range, 'wave_direction': [0, np.pi/2], 'radiating_dof': ['Heave']})
    dataset = solver.fill_dataset(test_matrix, [sphere])

    np.testing.assert_allclose(sorted(dataset.coords['period']), sorted(T_range))
    assert set(dataset.added_mass.dims) == {'period', 'radiating_dof', 'influenced_dof'}
    assert set(dataset.wavenumber.dims) == {'period'}
    assert set(dataset.wavelength.dims) == {'period'}
    assert set(dataset.omega.dims)      == {'period'}
    assert set(dataset.period.dims)     == {'period'}


def test_fill_dataset_with_wavenumbers_and_several_water_depths(sphere, solver):
    k_range = np.linspace(1.0, 3.0, 3)
    test_matrix = xr.Dataset(coords={
        'wavenumber': k_range, 'radiating_dof': ['Heave'], 'water_depth': [4.0, 6.0],
    })
    dataset = solver.fill_dataset(test_matrix, [sphere])

    np.testing.assert_allclose(dataset.coords['wavenumber'], k_range)
    assert set(dataset.added_mass.dims) == {'wavenumber', 'radiating_dof', 'influenced_dof', 'water_depth'}
    assert set(dataset.wavenumber.dims) == {'wavenumber'}
    assert set(dataset.wavelength.dims) == {'wavenumber'}
    assert set(dataset.omega.dims)      == {'wavenumber', 'water_depth'}
    assert set(dataset.period.dims)     == {'wavenumber', 'water_depth'}



"""Legacy interfaces to predefined meshes"""
# Copyright (C) 2017-2022 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>

import logging
import numpy as np

from capytaine.meshes.predefined import mesh_rectangle, mesh_parallelepiped
from capytaine.meshes.meshes import Mesh
from capytaine.bodies.bodies import FloatingBody

LOG = logging.getLogger(__name__)


class Rectangle(FloatingBody):
    """One-sided vertical rectangle (along y and z).

    By default, the normals are oriented in the positive y direction.

    Parameters
    ----------
    size : couple of floats, optional
        dimensions of the rectangle (width and height)
    resolution : couple of ints, optional
        number of faces along each of the two directions
    center : 3-ple of floats, optional
        position of the geometric center of the rectangle, default: (0, 0, 0)
    normal: 3-ple of floats, optional
        normal vector, default: along x axis
    translational_symmetry : bool, optional
        if True, use the translation symmetry to speed up the computations
    reflection_symmetry : bool, optional
        if True, use the reflection symmetry to speed up the computations
    name : string, optional
        a name for the body
    """

    def __init__(self, size=(5.0, 5.0), resolution=(5, 5),
                 center=(0, 0, 0), normal=(1, 0, 0),
                 translational_symmetry=False, reflection_symmetry=False, name=None):

        LOG.warning("Deprecation warning: The class Rectangle() is deprecated. "
                    "Please prefer the function capytaine.meshes.predefined.mesh_rectangle()")

        self.size = np.asarray(size, dtype=float)
        self.geometric_center = np.asarray(center, dtype=float)

        if name is None:
            name = f"rectangle_{next(Mesh._ids)}"

        mesh = mesh_rectangle(size=size, resolution=resolution, center=center, normal=normal,
                translation_symmetry=translational_symmetry, reflection_symmetry=reflection_symmetry,
                name=f"{name}_mesh")
        FloatingBody.__init__(self, mesh=mesh, name=name)


class RectangularParallelepiped(FloatingBody):
    """Six rectangles forming a parallelepiped.

    Parameters
    ----------
    size : 3-ple of floats, optional
        dimensions of the parallelepiped (width, thickness, height) for coordinates (x, y, z).
    resolution : 3-ple of ints, optional
        number of faces along the three directions
    center : 3-ple of floats, optional
        coordinates of the geometric center of the parallelepiped
    top: bool, optional
        whether or not to close the parallelepiped on the top
    bottom: bool, optional
        whether or not to close the parallelepiped on the bottom
    reflection_symmetry : bool, optional
        use xOz and yOz symmetry plane to generate the mesh
    translational_symmetry : bool, optional
        if True, use the translation symmetry in the x direction to speed up the computations.
        To use the translation symmetry in the y direction, create a x-symmetric body and then rotate it by pi/2.
    name : string, optional
        a name for the body
    """

    def __init__(self,
                 size=(1.0, 1.0, 1.0), resolution=(4, 4, 4),
                 center=(0, 0, 0),
                 top=True, bottom=True,
                 reflection_symmetry=False,
                 translational_symmetry=False,
                 name=None):

        LOG.warning("Deprecation warning: The class RectangularParallelepiped() is deprecated. "
                    "Please prefer the function capytaine.meshes.predefined.mesh_parallelepiped()")

        if name is None:
            name = f"rectangular_parallelepiped_{next(Mesh._ids)}"

        missing_sides = set()
        if not top: missing_sides.add("top")
        if not bottom: missing_sides.add("bottom")

        mesh = mesh_parallelepiped(size=size, resolution=resolution, center=center,
                missing_sides=missing_sides,
                translation_symmetry=translational_symmetry, reflection_symmetry=reflection_symmetry,
                name=f"{name}_mesh")

        FloatingBody.__init__(self, mesh=mesh, name=name)

class OpenRectangularParallelepiped(RectangularParallelepiped):
    def __init__(self, *args, **kwargs):
        RectangularParallelepiped.__init__(self, top=False, bottom=False, *args, **kwargs)
        # Kept mostly for legacy



import pytest

import numpy as np
import xarray as xr

import capytaine as cpt
from capytaine import __version__

#-----------------------------------------------------------------------------#
# Test indirect solver
#-----------------------------------------------------------------------------#
@pytest.fixture
def sphere():
    mesh = cpt.mesh_sphere(radius=1.0, resolution=(4, 4)).immersed_part()
    body = cpt.FloatingBody(mesh=mesh)
    body.add_translation_dof(direction=(1, 0, 0), name="Surge")
    return body


def test_exportable_settings():
    gf = cpt.Delhommeau(tabulation_nr=10, tabulation_nz=10,
                    tabulation_method="legacy", tabulation_nb_integration_points=50)
    assert gf.exportable_settings['green_function'] == 'Delhommeau'
    assert gf.exportable_settings['tabulation_nb_integration_points'] == 50
    assert gf.exportable_settings['tabulation_method'] == "legacy"
    assert gf.exportable_settings['finite_depth_prony_decomposition_method'] == 'fortran'

    gf2 = cpt.XieDelhommeau()
    assert gf2.exportable_settings['green_function'] == 'XieDelhommeau'

    engine = cpt.BasicMatrixEngine(matrix_cache_size=0)
    assert engine.exportable_settings['engine'] == 'BasicMatrixEngine'
    assert engine.exportable_settings['matrix_cache_size'] == 0
    assert engine.exportable_settings['linear_solver'] == 'lu_decomposition'

    solver = cpt.BEMSolver(green_function=gf, engine=engine)
    assert solver.exportable_settings['green_function'] == 'Delhommeau'
    assert solver.exportable_settings['tabulation_nb_integration_points'] == 50
    assert solver.exportable_settings['finite_depth_prony_decomposition_method'] == 'fortran'
    assert solver.exportable_settings['engine'] == 'BasicMatrixEngine'
    assert solver.exportable_settings['matrix_cache_size'] == 0
    assert solver.exportable_settings['linear_solver'] == 'lu_decomposition'


def test_limit_frequencies(sphere):
    """Test if how the solver answers when asked for frequency of 0 or ∞."""
    solver = cpt.BEMSolver()

    solver.solve(cpt.RadiationProblem(body=sphere, omega=0.0, water_depth=np.inf))

    with pytest.raises(NotImplementedError):
        solver.solve(cpt.RadiationProblem(body=sphere, omega=0.0, water_depth=1.0))

    solver.solve(cpt.RadiationProblem(body=sphere, omega=np.inf, water_depth=np.inf))

    with pytest.raises(NotImplementedError):
        solver.solve(cpt.RadiationProblem(body=sphere, omega=np.inf, water_depth=10))


def test_limit_frequencies_with_symmetries():
    mesh = cpt.mesh_parallelepiped(reflection_symmetry=True).immersed_part()
    body = cpt.FloatingBody(mesh=mesh)
    body.add_translation_dof(name="Surge")
    pb = cpt.RadiationProblem(body=body, omega=0.0)
    solver = cpt.BEMSolver()
    res = solver.solve(pb, keep_details=True)
    assert isinstance(res.added_mass['Surge'], float)


def test_zero_frequency_datasets(sphere):
    test_matrix = xr.Dataset(coords={
        "wavenumber": np.linspace(0.0, 0.1, 3),
        "radiating_dof": list(sphere.dofs),
        })
    solver = cpt.BEMSolver()
    solver.fill_dataset(test_matrix, sphere)


def test_infinite_frequency_datasets(sphere):
    test_matrix = xr.Dataset(coords={
        "wavelength": np.linspace(0.0, 0.1, 3),
        "radiating_dof": list(sphere.dofs),
        })
    solver = cpt.BEMSolver()
    solver.fill_dataset(test_matrix, sphere)


def test_parallelization(sphere):
    joblib = pytest.importorskip("joblib")
    solver = cpt.BEMSolver()
    test_matrix = xr.Dataset(coords={
        'omega': np.linspace(0.1, 4.0, 3),
        'radiating_dof': list(sphere.dofs.keys()),
    })
    solver.fill_dataset(test_matrix, sphere, n_jobs=2)


def test_float32_solver(sphere):
    solver = cpt.BEMSolver(green_function=cpt.Delhommeau(floating_point_precision="float32"))
    pb = cpt.RadiationProblem(body=sphere, radiating_dof="Surge", omega=1.0)
    solver.solve(pb)


def test_fill_dataset(sphere):
    solver = cpt.BEMSolver()
    test_matrix = xr.Dataset(coords={
        'omega': np.linspace(0.1, 4.0, 3),
        'wave_direction': np.linspace(0.0, np.pi, 3),
        'radiating_dof': list(sphere.dofs.keys()),
        'rho': [1025.0],
        'water_depth': [np.inf, 10.0],
        'g': [9.81]
    })
    dataset = solver.fill_dataset(test_matrix, sphere, n_jobs=1)

    # Tests on the coordinates
    assert list(dataset.coords['influenced_dof']) == list(dataset.coords['radiating_dof']) == list(sphere.dofs.keys())
    assert dataset.body_name == sphere.name
    assert dataset.rho == test_matrix.rho
    assert dataset.g == test_matrix.g

    # Tests on the results
    assert 'added_mass' in dataset
    assert 'radiation_damping' in dataset
    assert 'Froude_Krylov_force' in dataset
    assert 'diffraction_force' in dataset

    # Test the attributes
    assert dataset.attrs['capytaine_version'] == __version__
    assert 'start_of_computation' in dataset.attrs

    # Try to strip out the outputs and recompute
    naked_data = dataset.drop_vars(["added_mass", "radiation_damping", "diffraction_force", "Froude_Krylov_force"])
    recomputed_dataset = solver.fill_dataset(naked_data, [sphere])
    assert recomputed_dataset.rho == dataset.rho
    assert recomputed_dataset.g == dataset.g
    assert "added_mass" in recomputed_dataset
    assert np.allclose(recomputed_dataset["added_mass"].data, dataset["added_mass"].data)


# TODO: move the code below to test_io_xarray.py
    # wavenumbers = wavenumber_data_array(results)
    # assert isinstance(wavenumbers, xr.DataArray)

#-----------------------------------------------------------------------------#
# Test direct solver
#-----------------------------------------------------------------------------#
def test_direct_solver(sphere):
    problem = cpt.DiffractionProblem(body=sphere, omega=1.0)
    solver = cpt.BEMSolver()
    direct_result = solver.solve(problem, method='direct')
    indirect_result = solver.solve(problem, method='indirect')
    assert direct_result.forces["Surge"] == pytest.approx(indirect_result.forces["Surge"], rel=1e-1)
